import time

import tooltable  # type: ignore
import tooltime
import toolstr

from ctc import directory
from ctc import evm
from ctc import rpc
from ctc.protocols import chainlink_utils


async def async_summarize_feed(feed, n_recent=10):

    if evm.is_address_str(feed):
        feed_address = feed
    elif isinstance(feed, str):
        feed_address = directory.get_oracle_address(
            name=feed, protocol='chainlink'
        )
    else:
        raise Exception('unknown feed specification: ' + str(feed))

    call_kwargs = {'to_address': feed_address}
    name = await rpc.async_eth_call(function_name='description', **call_kwargs)
    decimals = await rpc.async_eth_call(function_name='decimals', **call_kwargs)
    aggregator = await rpc.async_eth_call(
        function_name='aggregator', **call_kwargs
    )

    latest_block = await rpc.async_eth_block_number()
    data = await chainlink_utils.async_get_feed_data(
        feed_address,
        fields='full',
        start_block=latest_block - 5000,
    )
    most_recent = data.iloc[-1]
    timestamp = most_recent['timestamp']

    updates = []
    for i, row in data.iterrows():
        timestamp = row['timestamp']
        age = tooltime.timelength_to_phrase(time.time() - timestamp)
        age = ' '.join(age.split(' ')[:2]).strip(',')
        update = [
            row['answer'],
            age,
            str(i),
            str(timestamp),
            tooltime.timestamp_to_iso(timestamp).replace('T', ' '),
        ]
        updates.append(update)

    title = 'Chainlink Feed Summary: ' + name
    toolstr.print_text_box(title, double=False)
    print('- name:', name)
    print('- address:', feed_address)
    print('- aggregator:', aggregator)
    print('- decimals:', decimals)
    # print('- deviation threshold:')
    # print('- heartbeat:')
    print()
    print()
    toolstr.print_header('Recent Updates')
    print()
    headers = ['value', 'age', 'block', 'timestamp', 'time']
    tooltable.print_table(
        updates[-n_recent:][::-1], headers=headers, indent='    '
    )




from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_eth_get_work(provider=None):
    request = rpc_constructors.construct_eth_get_work()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_work(response=response)


async def async_eth_submit_work(nonce, pow_hash, digest, provider=None):
    request = rpc_constructors.construct_eth_submit_work(
        nonce=nonce,
        pow_hash=pow_hash,
        digest=digest,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_submit_work(response=response)


async def async_eth_submit_hashrate(hashrate, id, provider=None):
    request = rpc_constructors.construct_eth_submit_hashrate(
        hashrate=hashrate,
        id=id,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_submit_hashrate(response=response)


async def async_eth_coinbase(provider=None):
    request = rpc_constructors.construct_eth_coinbase()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_coinbase(response=response)


async def async_eth_mining(provider=None):
    request = rpc_constructors.construct_eth_mining()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_mining(response=response)


async def async_eth_hashrate(provider=None):
    request = rpc_constructors.construct_eth_hashrate()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_hashrate(response=response)




import asyncio

from ctc import evm
from ctc import rpc
from . import coracle_spec


async def async_get_pcv_stats(
    block=None, blocks=None, wrapper=False, provider=None
):

    if block is not None:
        block = await evm.async_block_number_to_int(block=block)
    if blocks is not None:
        blocks = await evm.async_block_numbers_to_int(blocks=blocks)

    # assemble kwargs
    provider = rpc.get_provider(provider)
    if provider['chunk_size'] is None:
        provider['chunk_size'] = 1

    # fetch results
    keys = ['pcv', 'user_fei', 'protocol_equity', 'valid']
    if block is not None or (block is None and blocks is None):
        to_address = coracle_spec.get_coracle_address(
            wrapper, block=block,
        )
        result = await rpc.async_eth_call(
            function_name='pcvStats',
            block_number=block,
            provider=provider,
            to_address=to_address,
        )
        return dict(zip(keys, result))

    elif blocks is not None:
        coroutines = []
        for block in blocks:
            to_address = coracle_spec.get_coracle_address(
                wrapper, block=block,
            )
            coroutine = rpc.async_eth_call(
                function_name='pcvStats',
                block_number=block,
                provider=provider,
                to_address=to_address,
            )
            coroutines.append(coroutine)
        result = await asyncio.gather(*coroutines)

        # arrange results
        transpose = list(zip(*result))
        data = {}
        for k, key in enumerate(keys):
            data[key] = transpose[k]
            import numpy as np

            data[key] = np.array(data[key])
            if key in ['pcv', 'user_fei', 'protocol_equity']:
                data[key] = data[key] / 1e18

        # create dataframe
        import pandas as pd

        df = pd.DataFrame(data, index=blocks)

        return df

    else:
        raise Exception('must specify block or blocks')




from __future__ import annotations

import asyncio

from ctc import evm
from ctc import spec
from ctc.protocols import fei_utils
from .. import analytics_spec


async def async_compute_pcv_stats(
    blocks: list[int],
    verbose: bool = False,
) -> dict[str, analytics_spec.MetricGroup]:

    pcv_stats_task = asyncio.create_task(
        fei_utils.async_get_pcv_stats(blocks=blocks)
    )
    total_supply_task = asyncio.create_task(
        evm.async_get_erc20_total_supply_by_block(token='FEI', blocks=blocks)
    )

    pcv_stats = await pcv_stats_task

    user_fei = pcv_stats['user_fei']
    pcv_total = pcv_stats['pcv']
    cr = pcv_stats['pcv'] / user_fei

    total_supply = await total_supply_task
    protocol_fei = [
        block_total_supply - block_user_fei
        for block_total_supply, block_user_fei in zip(total_supply, user_fei)
    ]

    return {
        'pcv_stats': {
            'name': 'PCV Statistics',
            'metrics': {
                'pcv_total': {
                    'name': 'PCV Total',
                    'values': list(pcv_total.values),
                },
                'cr': {
                    'name': 'Collateralization Ratio',
                    'values': list(cr.values),
                },
            },
        },
        'circulating_fei': {
            'name': 'Circulating FEI',
            'metrics': {
                'user_fei': {
                    'name': 'User FEI',
                    'values': list(user_fei.values),
                },
                'protocol_fei': {
                    'name': 'Protocol FEI',
                    'values': protocol_fei,
                },
            },
        },
    }


async def async_compute_pcv_by_asset(
    blocks: list[int], verbose: bool = False, provider: spec.ProviderSpec = None
) -> analytics_spec.MetricGroup:

    tokens_balances = await fei_utils.async_get_tokens_balances_by_block(
        blocks=blocks,
        usd=True,
        exclude_fei=True,
    )

    tokens = list(tokens_balances.keys())

    symbols = await fei_utils.async_get_pcv_tokens_symbols(
        tokens=tokens,
        provider=provider,
        block=blocks[-1],
    )

    metrics: dict[str, analytics_spec.MetricData] = {}
    for token in tokens:
        symbol = symbols[token]

        if symbol in fei_utils.token_aliases:
            symbol = fei_utils.token_aliases[symbol]

        metrics[symbol] = {
            'name': symbol,
            'values': tokens_balances[token],
            'units': 'USD',
        }

    return {'name': 'PCV by Asset', 'metrics': metrics}




from ctc import binary


def is_address_str(some_str):
    return (
        isinstance(some_str, str)
        and some_str.startswith('0x')
        and len(some_str) == 42
    )


def get_created_address(
    sender, nonce=None, salt=None, init_code=None, output_format=None
):
    # see https://github.com/ethereum/EIPs/blob/master/EIPS/eip-1014.md
    # see https://ethereum.stackexchange.com/a/761
    if nonce is not None:
        # create
        sender = binary.convert(sender, 'binary')
        data = binary.rlp_encode([sender, nonce])
    elif salt is not None and init_code is not None:
        # create2
        data = (
            binary.convert('0xff', 'raw_hex')
            + binary.convert(sender, 'raw_hex')
            + binary.convert(salt, 'raw_hex')
            + binary.convert(binary.keccak(init_code), 'raw_hex')
        )
    else:
        raise Exception('specify either {nonce} or {salt, init_code}')

    result = binary.keccak(data, output_format='prefix_hex')
    result = '0x' + result[26:]
    if output_format is not None and output_format != 'prefix_hex':
        result = binary.keccak(data=result, output_format=output_format)

    return result


def create_hash_preview(
    hash_data,
    show_start=True,
    show_end=True,
    include_0x=True,
    n_chars=6,
    n_chars_start=None,
    n_chars_end=None,
):

    if not include_0x and hash_data[:2] == '0x':
        hash_data = hash_data[2:]

    if n_chars_start is None:
        n_chars_start = n_chars
        if include_0x:
            n_chars_start += 2
    if n_chars_end is None:
        n_chars_end = n_chars

    preview = '...'
    if show_start:
        preview = hash_data[:n_chars_start] + preview
    if show_end:
        preview = preview + hash_data[:-n_chars_end]

    return preview


def get_address_checksum(address):
    """

    - adapted from eth_utils.to_checksum_address()
    """

    # validate address
    address_format = binary.get_binary_format(address)
    if address_format not in ['prefix_hex', 'raw_hex']:
        raise Exception('checksum only relevant to hex formatted addresses')
    address = binary.convert(address, 'raw_hex')

    # compute address hash
    address_hash = binary.keccak_text(
        address.lower(),
        output_format='raw_hex',
    )

    # assemble checksum
    chars = []
    for address_char, hash_char in zip(address, address_hash):
        if int(hash_char, 16) > 7:
            chars.append(address_char.upper())
        else:
            chars.append(address_char.lower())
    raw_checksum = ''.join(chars)

    # convert to output format
    if address_format == 'raw_hex':
        return raw_checksum
    elif address_format == 'prefix_hex':
        return binary.convert(raw_checksum, 'prefix_hex')
    else:
        raise Exception('checksum only relevant to hex formatted addresses')


def create_reverse_address_map(
    address_map, include_lower=True, include_checksum=True
):
    if not include_lower and not include_checksum:
        raise Exception(
            'must include lower case addresses and/or checksum addresses'
        )

    reverse_map = {}
    for name, address in address_map.items():
        if include_lower:
            reverse_map[address.lower()] = name
        if include_checksum:
            reverse_map[get_address_checksum(address)] = name

    return reverse_map




import pytest
import numpy as np

from ctc.toolbox.amm_utils import cpmm
from ctc.toolbox import validate_utils


x_reserves = 1e6
y_reserves = 1e3


@pytest.mark.parametrize('x_sold', [1e1, 1e6, 1e10])
def test_trade_x_sold(x_sold):

    result = cpmm.trade(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        x_sold=x_sold,
    )

    assert result['new_pool']['x_reserves'] == x_reserves + x_sold
    assert result['new_pool']['y_reserves'] == y_reserves - result['y_bought']
    assert result['y_bought'] == cpmm.compute_y_bought_when_x_sold(
        x_sold=x_sold,
        x_reserves=x_reserves,
        y_reserves=y_reserves,
    )


@pytest.mark.parametrize('y_sold', [1e1, 1e6, 1e10])
def test_trade_y_sold(y_sold):

    result = cpmm.trade(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        y_sold=y_sold,
    )

    assert result['new_pool']['x_reserves'] == x_reserves - result['x_bought']
    assert result['new_pool']['y_reserves'] == y_reserves + y_sold
    assert result['x_bought'] == cpmm.compute_y_bought_when_x_sold(
        x_sold=y_sold,
        x_reserves=y_reserves,
        y_reserves=x_reserves,
    )


@pytest.mark.parametrize('x_bought', [1e1, 1e3, 1e5])
def test_trade_x_bought(x_bought):

    result = cpmm.trade(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        x_bought=x_bought,
    )
    assert result['new_pool']['x_reserves'] == x_reserves - x_bought
    assert result['new_pool']['y_reserves'] == y_reserves + result['y_sold']

    # ensure that the reserve specification gives the same result
    reverse_result = cpmm.trade(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        y_sold=result['y_sold'],
    )
    validate_utils._ensure_values_equal(result, reverse_result)


@pytest.mark.parametrize('y_bought', [1e0, 1e1, 1e2])
def test_trade_y_bought(y_bought):

    result = cpmm.trade(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        y_bought=y_bought,
    )
    assert result['new_pool']['x_reserves'] == x_reserves - result['x_bought']
    assert result['new_pool']['y_reserves'] == y_reserves - y_bought

    # ensure that the reserve specification gives the same result
    reverse_result = cpmm.trade(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        x_sold=result['x_sold'],
    )
    validate_utils._ensure_values_equal(result, reverse_result)


@pytest.mark.parametrize('new_x_reserves', [1e1, 1e6, 1e9])
def test_trade_new_x_reserves(new_x_reserves):

    result = cpmm.trade_to_target_reserves(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        new_x_reserves=new_x_reserves,
    )
    assert result['new_pool']['x_reserves'] == new_x_reserves

    # ensure that the resulting trade produces the same results
    count = 0
    for name in ['x_sold', 'y_sold', 'y_bought', 'x_bought']:
        if result[name] >= 0:
            reverse_result = cpmm.trade(
                x_reserves=x_reserves,
                y_reserves=y_reserves,
                **{name: result[name]}
            )
            validate_utils._ensure_values_equal(result, reverse_result)
            count += 1
    if count < 2:
        raise Exception('could not detect enough suitable reverse trades')


@pytest.mark.parametrize('new_y_reserves', [1e1, 1e3, 1e6])
def test_trade_y_reserves_new(new_y_reserves):

    result = cpmm.trade_to_target_reserves(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        new_y_reserves=new_y_reserves,
    )
    assert result['new_pool']['y_reserves'] == new_y_reserves

    # ensure that the resulting trade produces the same results
    count = 0
    for name in ['x_sold', 'y_sold', 'y_bought', 'x_bought']:
        if result[name] >= 0:
            reverse_result = cpmm.trade(
                x_reserves=x_reserves,
                y_reserves=y_reserves,
                **{name: result[name]}
            )
            validate_utils._ensure_values_equal(result, reverse_result)
            count += 1
    if count < 2:
        raise Exception('could not detect enough suitable reverse trades')


@pytest.mark.parametrize('new_price', [1e-8, 1e-2, 1e2, 1e8])
def test_trade_to_price(new_price):

    # modify x per y
    result = cpmm.trade_to_price(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        new_x_per_y=new_price,
    )
    result_price = (
        result['new_pool']['x_reserves'] / result['new_pool']['y_reserves']
    )
    assert np.isclose(new_price, result_price)

    # modify y per x
    result = cpmm.trade_to_price(
        x_reserves=x_reserves,
        y_reserves=y_reserves,
        new_y_per_x=new_price,
    )
    result_price = (
        result['new_pool']['y_reserves'] / result['new_pool']['x_reserves']
    )
    assert np.isclose(new_price, result_price)


def test_reject_negative_values():
    for arg in [
        'x_sold',
        'y_sold',
        'x_bought',
        'y_bought',
    ]:
        with pytest.raises(Exception):
            cpmm.trade(
                x_reserves=x_reserves, y_reserves=y_reserves, **{arg: -1}
            )




filesystem_layout = {
    'evm_events_path': 'events/contract__{contract_address}/event__{event_hash}/{start_block}__to__{end_block}.csv',
    'evm_contract_abis_path': 'contract_abis/contract__{contract_address}/{name}.json',
    'evm_named_contract_abis_path': '{data_root}/{network}/evm/named_contract_abis',
}


erc20_abis = {
    'name': {
        'constant': True,
        'inputs': [],
        'name': 'name',
        'outputs': [{'name': '', 'type': 'string'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'approve': {
        'constant': False,
        'inputs': [
            {'name': '_spender', 'type': 'address'},
            {'name': '_value', 'type': 'uint256'},
        ],
        'name': 'approve',
        'outputs': [{'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'totalSupply': {
        'constant': True,
        'inputs': [],
        'name': 'totalSupply',
        'outputs': [{'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'transferFrom': {
        'constant': False,
        'inputs': [
            {'name': '_from', 'type': 'address'},
            {'name': '_to', 'type': 'address'},
            {'name': '_value', 'type': 'uint256'},
        ],
        'name': 'transferFrom',
        'outputs': [{'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'decimals': {
        'constant': True,
        'inputs': [],
        'name': 'decimals',
        'outputs': [{'name': '', 'type': 'uint8'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'balanceOf': {
        'constant': True,
        'inputs': [{'name': '_owner', 'type': 'address'}],
        'name': 'balanceOf',
        'outputs': [{'name': 'balance', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'symbol': {
        'constant': True,
        'inputs': [],
        'name': 'symbol',
        'outputs': [{'name': '', 'type': 'string'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'transfer': {
        'constant': False,
        'inputs': [
            {'name': '_to', 'type': 'address'},
            {'name': '_value', 'type': 'uint256'},
        ],
        'name': 'transfer',
        'outputs': [{'name': '', 'type': 'bool'}],
        'payable': False,
        'stateMutability': 'nonpayable',
        'type': 'function',
    },
    'allowance': {
        'constant': True,
        'inputs': [
            {'name': '_owner', 'type': 'address'},
            {'name': '_spender', 'type': 'address'},
        ],
        'name': 'allowance',
        'outputs': [{'name': '', 'type': 'uint256'}],
        'payable': False,
        'stateMutability': 'view',
        'type': 'function',
    },
    'Approval': {
        'anonymous': False,
        'inputs': [
            {'indexed': True, 'name': 'owner', 'type': 'address'},
            {'indexed': True, 'name': 'spender', 'type': 'address'},
            {'indexed': False, 'name': 'value', 'type': 'uint256'},
        ],
        'name': 'Approval',
        'type': 'event',
    },
    'Transfer': {
        'anonymous': False,
        'inputs': [
            {'indexed': True, 'name': 'from', 'type': 'address'},
            {'indexed': True, 'name': 'to', 'type': 'address'},
            {'indexed': False, 'name': 'value', 'type': 'uint256'},
        ],
        'name': 'Transfer',
        'type': 'event',
    },
}




import typing

from ctc import spec
from . import timestamp_crud
from . import metric_crud
from . import analytics_spec


async def async_create_payload(
    *,
    blocks: typing.Sequence[spec.BlockNumberReference] = None,
    timestamps: typing.Sequence[int] = None,
    timescale: typing.Optional[analytics_spec.TimescaleSpec] = None,
    end_time: typing.Optional[analytics_spec.Timestamp] = None,
    window_size: str = None,
    interval_size: str = None,
    provider: spec.ProviderSpec = None,
) -> analytics_spec.AnalyticsPayload:
    """create data payload from scratch"""

    time_data = await timestamp_crud.async_get_time_data(
        blocks=blocks,
        timestamps=timestamps,
        timescale=timescale,
        end_time=end_time,
        window_size=window_size,
        interval_size=interval_size,
        provider=provider,
    )

    # get data
    data = await metric_crud.async_get_metrics(
        blocks=time_data['block_numbers']
    )

    return {
        'version': '0.1.0',
        #
        # time data
        'n_samples': time_data['n_samples'],
        'window_size': time_data['window_size'],
        'interval_size': time_data['interval_size'],
        'timestamps': time_data['timestamps'],
        'block_numbers': time_data['block_numbers'],
        'created_at_timestamp': time_data['created_at_timestamp'],
        #
        # metric data
        'data': data,
    }


# def update_payload(
#     timescale: analytics_spec.Timescale,
#     old_payload: analytics_spec.AnalyticsPayload,
# ) -> analytics_spec.AnalyticsPayload:

#     new_timestamps = get_new_timestamps(
#         timescale=timescale,
#         old_payload=old_payload,
#     )
#     new_blocks = get_new_blocks(
#         new_timestamps=new_timestamps,
#         old_payload=old_payload,
#     )
#     new_metrics = get_metrics(blocks=new_blocks)

#     return combine_new_data(
#         old_payload=old_payload,
#         new_metrics=new_metrics,
#     )




from __future__ import annotations

import typing

from ctc import binary
from ctc import spec

from . import contract_parsing
from . import function_parsing


_Parameters = typing.Union[
    typing.Sequence[typing.Any],
    typing.Mapping[str, typing.Any],
]


#
# # call data
#


def encode_call_data(
    *,
    function_selector: typing.Optional[str] = None,
    parameter_types: typing.Optional[spec.ABIDatumType] = None,
    parameters: typing.Optional[_Parameters] = None,
    encoded_parameters: typing.Optional[spec.BinaryData] = None,
    function_abi: typing.Optional[spec.FunctionABI] = None,
) -> str:

    # encode function selector
    if function_selector is None:
        function_selector = function_parsing.get_function_selector(function_abi)
    function_selector = binary.convert(function_selector, 'prefix_hex')

    # encode parameters
    if encoded_parameters is None:
        encoded_parameters = encode_function_parameters(
            parameters=parameters,
            parameter_types=parameter_types,
            function_abi=function_abi,
        )
    encoded_parameters = binary.convert(encoded_parameters, 'raw_hex')

    # join function selector with parameters
    return function_selector + encoded_parameters


def decode_call_data(
    call_data: spec.BinaryData,
    function_abi: typing.Optional[spec.FunctionABI] = None,
    contract_abi: typing.Optional[spec.ContractABI] = None,
) -> spec.DecodedCallData:

    # get function selector
    call_data_bytes = binary.convert(call_data, 'binary')
    function_selector = binary.convert(call_data_bytes[:4], 'prefix_hex')

    # get function abi
    if function_abi is None:
        if contract_abi is None:
            raise Exception('must specify function_abi or contract_abi')
        function_abi = contract_parsing.get_function_abi(
            contract_abi=contract_abi,
            function_selector=function_selector,
        )

    # decode parameters
    encoded_parameters = call_data_bytes[4:]
    parameter_types = function_parsing.get_function_parameter_types(
        function_abi
    )
    decoded_parameters = decode_function_parameters(
        encoded_parameters, parameter_types
    )

    # compute named parameters
    parameter_names = function_parsing.get_function_parameter_names(
        function_abi
    )
    if len(parameter_names) == len(decoded_parameters) and all(
        name is not None for name in parameter_names
    ):
        named_parameters = typing.cast(
            dict[str, typing.Any],
            dict(zip(parameter_names, decoded_parameters)),
        )
    else:
        named_parameters = None

    return {
        'function_abi': function_abi,
        'function_selector': function_selector,
        'parameters': decoded_parameters,
        'named_parameters': named_parameters,
    }


#
# # function parameters
#


def encode_function_parameters(
    parameters: typing.Optional[_Parameters] = None,
    parameter_types: typing.Optional[typing.Sequence[spec.ABIDatumType]] = None,
    function_signature: typing.Optional[str] = None,
    function_abi: typing.Optional[spec.FunctionABI] = None,
) -> bytes:

    if parameters is None:
        return bytes()

    # get parameter types
    if parameter_types is None:
        parameter_types = function_parsing.get_function_parameter_types(
            function_signature=function_signature,
            function_abi=function_abi,
        )

    # convert parameter dict to list
    if isinstance(parameters, typing.Mapping):
        if function_abi is None:
            raise Exception('must specify function_abi')
        parameter_names = function_parsing.get_function_parameter_names(
            function_abi=function_abi,
            require_names=True,
        )
        parameters = [parameters[name] for name in parameter_names]

    # inefficient: convert prefix_hex binary to bytes
    new_parameters = []
    for parameter_type, parameter in zip(parameter_types, parameters):
        if (
            parameter_type == 'bytes32'
            and binary.get_binary_format(parameter) != 'binary'
        ):
            parameter = binary.convert(parameter, 'binary')
        new_parameters.append(parameter)
    parameters = new_parameters

    # encode
    encoded_bytes = binary.abi_encode(
        parameters, '(' + ','.join(parameter_types) + ')'
    )

    # convert to output format
    return encoded_bytes


def decode_function_parameters(
    encoded_parameters: spec.BinaryData,
    parameter_types: list[spec.ABIDatumType],
) -> list[typing.Any]:

    parameter_types_str = '(' + ','.join(parameter_types) + ')'
    encoded_parameters = binary.convert(encoded_parameters, 'binary')
    parameters = binary.abi_decode(encoded_parameters, parameter_types_str)

    return parameters


def decode_function_named_parameters(
    function_abi: spec.FunctionABI,
    encoded_parameters: spec.BinaryData,
    parameter_types: typing.Optional[list[spec.ABIDatumType]] = None,
) -> dict[str, typing.Any]:

    if parameter_types is None:
        parameter_types = function_parsing.get_function_parameter_types(
            function_abi=function_abi,
        )

    decoded_parameters = decode_function_parameters(
        encoded_parameters=encoded_parameters,
        parameter_types=parameter_types,
    )

    # get parameter names
    parameter_names = function_parsing.get_function_parameter_names(
        function_abi
    )

    return {
        parameter_name: parameter
        for parameter_name, parameter in zip(
            parameter_names, decoded_parameters
        )
        if parameter_name is not None
    }


def decode_function_output(
    encoded_output: spec.BinaryData,
    output_types: typing.Optional[list[spec.ABIDatumType]] = None,
    delist_single_outputs: bool = True,
    package_named_outputs: bool = False,
    function_abi: spec.FunctionABI = None,
) -> typing.Any:
    # need to test case when function has no output

    # get output types
    if output_types is None:
        if function_abi is None:
            raise Exception('must specify function_abi')
        output_types = function_parsing.get_function_output_types(function_abi)
    output_types_str = '(' + ','.join(output_types) + ')'

    # decode
    encoded_output = binary.convert(encoded_output, 'binary')
    decoded_output = binary.abi_decode(encoded_output, output_types_str)

    # decode strings
    new_decoded_output = []
    for output_type, item in zip(output_types, decoded_output):
        if output_type == 'string':
            # item = item.decode()
            item = item
        elif output_type == 'bytes32':
            item = binary.convert(item, 'prefix_hex')
        new_decoded_output.append(item)
    decoded_output = new_decoded_output

    # delist
    if delist_single_outputs and len(output_types) == 1:
        decoded_output = decoded_output[0]

    # repackage
    elif package_named_outputs and len(output_types) > 1:
        if function_abi is None:
            raise Exception('must specify function_abi')
        names = function_parsing.get_function_output_names(function_abi)
        if all(name is not None for name in names):
            decoded_output = dict(zip(names, decoded_output))

    return decoded_output




# replace eth_abi with eth_abi_lite

import typing

from ctc import spec


def abi_decode(data: bytes, types: spec.ABIDatatypeStr) -> typing.Any:
    import eth_abi

    return eth_abi.decode_single(types, data)


def abi_encode(data: typing.Any, types: spec.ABIDatatypeStr) -> bytes:
    import eth_abi

    return eth_abi.encode_single(types, data)




import math


def chunk_blocks_by_size(blocks, n_per_chunk):
    n_chunks = math.ceil(len(blocks) / n_per_chunk)
    return [
        blocks[c * n_per_chunk: (c + 1) * n_per_chunk]
        for c in range(n_chunks)
    ]


def chunk_blocks_into_ranges(blocks, chunk_size):
    chunks_by_group = {}
    for block in blocks:
        group = math.floor(block / chunk_size) * chunk_size
        chunks_by_group.setdefault(group, [])
        chunks_by_group[group].append(block)
    return chunks_by_group


def get_chunks_in_range(start_block, end_block, chunk_size, trim_excess=False):
    """break a range of blocks into chunks of a given chunk size"""
    chunk_start_block = (start_block // chunk_size) * chunk_size
    chunk_end_block = ((end_block // chunk_size) + 1) * chunk_size
    chunk_bounds = list(
        range(chunk_start_block, chunk_end_block + chunk_size, chunk_size)
    )
    chunks = [
        [start, end - 1]
        for start, end in zip(chunk_bounds[:-1], chunk_bounds[1:])
    ]

    if trim_excess:
        if len(chunks) > 0:
            if chunks[0][0] < start_block:
                chunks[0] = [start_block, chunks[0][1]]
            if chunks[-1][-1] > end_block:
                chunks[-1] = [chunks[-1][0], end_block]

    return chunks




from .rpc_backends import *
from .rpc_batch import *
from .rpc_constructors import *
from .rpc_digestors import *
from .rpc_executors import *
from .rpc_executors_async import *

from .rpc_format import *
from .rpc_lifecycle import *
from .rpc_provider import *
from .rpc_registry import *
from .rpc_request import *
from .rpc_spec import *




import typing

from ctc import spec
from ctc import binary


def digest_eth_call(
    response: spec.RpcSingularResponse,
    function_abi: spec.FunctionABI,
    decode_response: bool = True,
    delist_single_outputs: bool = True,
    package_named_outputs: bool = False,
    fill_empty: bool = False,
    empty_token: typing.Any = None,
) -> spec.RpcSingularResponse:

    if decode_response:

        if (fill_empty or empty_token is not None) and response == '0x':
            response = empty_token

        else:
            response = binary.decode_function_output(
                encoded_output=response,
                delist_single_outputs=delist_single_outputs,
                package_named_outputs=package_named_outputs,
                function_abi=function_abi,
            )

    return response


def digest_eth_estimate_gas(
    response: spec.RpcSingularResponse, decode_response: bool = True
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_get_balance(
    response: spec.RpcSingularResponse, decode_response: bool = True
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_get_storage_at(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_get_code(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response




import os

import toolcli


def output_data(data, output, overwrite, top=None, indent=None, raw=False):

    import pandas as pd

    if output == 'stdout':
        # print(data)
        import tooltable
        import toolstr

        rows = []
        if isinstance(data, pd.DataFrame):
            iterator = data.iterrows()
        elif isinstance(data, pd.Series):
            iterator = data.iteritems()
        else:
            raise Exception('unknown data format')

        for index, values in iterator:
            row = []
            row.append(index)
            if hasattr(values, 'values'):
                # dataframe
                for value in values.values:
                    if value and not isinstance(value, str):
                        value = toolstr.format(value)
                    row.append(value)
            else:
                # series
                if raw:
                    row.append(values)
                else:
                    row.append(toolstr.format(values, order_of_magnitude=True))
            rows.append(row)

        if top is not None:
            if len(rows) > top:
                rows = rows[:top]

        if isinstance(data, pd.DataFrame):
            columns = [data.index.name] + list(data.columns)
        elif isinstance(data, pd.Series):
            columns = [data.index.name, data.name]
        else:
            raise Exception('unknown data format')
        tooltable.print_table(rows=rows, headers=columns, indent=indent)

    else:

        # check whether file exists
        if os.path.isfile(output):
            if overwrite:
                pass
            elif toolcli.input_yes_or_no('File already exists. Overwrite? '):
                pass
            else:
                raise Exception('aborting')

        if output.endswith('.csv'):
            data.to_csv(output)

        elif output.endswith('.json'):
            data.to_json(output)

        else:
            raise Exception('unknown output format: ' + str(output))




from __future__ import annotations

import asyncio
import typing

from ctc import directory
from ctc import evm
from ctc import rpc
from ctc import spec
from ctc.toolbox import async_utils
from ctc.toolbox import nested_utils
from . import coracle_tokens
from . import coracle_oracles
from . import coracle_deposits
from . import coracle_spec


#
# # deposit balance getters
#


async def async_get_deposit_balance(
    deposit: spec.ContractAddress,
    block: spec.BlockReference = 'latest',
    provider: spec.ProviderSpec = None,
) -> typing.Union[int, list[int]]:
    """get token balance of a particular deposit"""
    return await rpc.async_eth_call(
        to_address=deposit,
        function_name='balance',
        block_number=block,
        provider=provider,
    )


async def async_get_deposits_balances(
    deposits: typing.Sequence[spec.ContractAddress],
    block: spec.BlockReference = 'latest',
    provider: spec.ProviderSpec = None,
) -> list[int]:
    return await rpc.async_batch_eth_call(
        to_addresses=deposits,
        function_name='balance',
        block_number=block,
        provider=provider,
    )


async def async_get_deposit_balance_by_block(
    deposit: spec.ContractAddress,
    blocks: typing.Sequence[spec.BlockReference],
    provider: spec.ProviderSpec = None,
) -> list[int]:
    return await rpc.async_batch_eth_call(
        to_address=deposit,
        function_name='balance',
        block_numbers=blocks,
        provider=provider,
    )


async def async_get_deposit_resistant_balance_and_fei(
    deposit: spec.ContractAddress,
    block: spec.BlockReference = 'latest',
    provider: spec.ProviderSpec = None,
) -> typing.Union[int, list[int]]:
    """get token balance of a particular deposit"""
    return await rpc.async_eth_call(
        to_address=deposit,
        function_name='resistantBalanceAndFei',
        block_number=block,
        provider=provider,
    )


async def async_get_deposits_resistant_balances_and_fei(
    deposits: typing.Sequence[spec.ContractAddress],
    block: spec.BlockReference = 'latest',
    provider: spec.ProviderSpec = None,
) -> list[int]:
    return await rpc.async_batch_eth_call(
        to_addresses=deposits,
        function_name='resistantBalanceAndFei',
        block_number=block,
        provider=provider,
    )


async def async_get_deposit_resistant_balance_and_fei_by_block(
    deposit: spec.ContractAddress,
    blocks: typing.Sequence[spec.BlockReference],
    provider: spec.ProviderSpec = None,
) -> list[int]:
    return await rpc.async_batch_eth_call(
        to_address=deposit,
        function_name='resistantBalanceAndFei',
        block_numbers=blocks,
        provider=provider,
    )


#
# # token balance getters
#


async def async_get_token_balance(
    token: spec.Address,
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
    usd: bool = False,
) -> typing.Union[int, float]:

    block = await evm.async_block_number_to_int(block=block, provider=provider)
    deposits = await coracle_deposits.async_get_token_deposits(
        token=token, block=block, provider=provider
    )
    balances = await async_get_deposits_balances(
        deposits=deposits, block=block, provider=provider
    )
    balance: typing.Union[int, float] = sum(balances)

    if normalize:
        if token == coracle_spec.usd_token:
            balance = balance / 1e18
        elif balance == 0:
            pass
        else:
            balance = await evm.async_normalize_erc20_quantity(
                quantity=balance,
                provider=provider,
                token=token,
            )

    if usd:
        if not normalize:
            raise Exception('must normalize for usd conversion')
        token_price = await coracle_oracles.async_get_token_price(
            token=token,
            block=block,
            provider=provider,
            normalize=True,
        )
        balance = balance * token_price

    return balance


async def async_get_token_balance_by_block(
    token: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
    usd: bool = False,
) -> typing.Union[list[int], list[float]]:

    coroutines = [
        async_get_token_balance(
            token=token,
            provider=provider,
            normalize=normalize,
            block=block,
            usd=usd,
        )
        for block in blocks
    ]
    return await asyncio.gather(*coroutines)


async def async_get_tokens_balances(
    tokens: typing.Sequence[spec.Address] = None,
    block: spec.BlockNumberReference = 'latest',
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
    usd: bool = False,
    exclude_fei: bool = True,
) -> typing.Union[dict[spec.Address, int], dict[spec.Address, float]]:

    block = await evm.async_block_number_to_int(block=block, provider=provider)

    # use all tokens in pcv by default
    if tokens is None:
        tokens = await coracle_tokens.async_get_tokens_in_pcv(
            block=block, provider=provider
        )
    if exclude_fei:
        FEI = directory.get_erc20_address(symbol='FEI')
        tokens = [token for token in tokens if token != FEI]

    # get deposits for each token
    tokens_deposits = await coracle_deposits.async_get_tokens_deposits(
        tokens=tokens, block=block, provider=provider
    )

    # get deposit balances
    all_deposits = [
        deposit
        for token_deposits in tokens_deposits.values()
        for deposit in token_deposits
    ]
    all_balances = await async_get_deposits_balances(
        deposits=all_deposits,
        provider=provider,
        block=block,
    )
    all_balances_iter = iter(all_balances)
    deposits_balances: dict[str, list[int]] = {}
    for token, token_deposits in tokens_deposits.items():
        deposits_balances[token] = []
        for _ in token_deposits:
            deposits_balances[token].append(next(all_balances_iter))

    # sum deposits balances of each token
    tokens_balances: dict[str, typing.Union[int, float]] = {
        token: sum(token_balances)
        for token, token_balances in deposits_balances.items()
    }

    # normalize
    if normalize:

        # normalize only non-usd tokens with non-zero balance
        normalize_tokens = []
        for token in tokens:
            if token != coracle_spec.usd_token and tokens_balances[token] > 0:
                normalize_tokens.append(token)

        # normalize and replace
        normalized = await evm.async_normalize_erc20s_quantities(
            quantities=[tokens_balances[token] for token in normalize_tokens],
            tokens=normalize_tokens,
            block=block,
            provider=provider,
        )
        for token, normalized_balance in zip(normalize_tokens, normalized):
            tokens_balances[token] = normalized_balance

        # normalize usd balance
        if coracle_spec.usd_token in tokens:
            tokens_balances[coracle_spec.usd_token] /= 1e18

    if usd:
        if not normalize:
            raise Exception('must normalize for usd conversion')
        token_prices = await coracle_oracles.async_get_tokens_prices(
            tokens=tokens,
            block=block,
            provider=provider,
            normalize=True,
        )
        tokens_balances = {
            token: tokens_balances[token] * token_price
            for token, token_price in zip(tokens_balances, token_prices)
        }

    return tokens_balances


async def async_get_tokens_balances_by_block(
    blocks: typing.Sequence[spec.BlockNumberReference],
    tokens: typing.Optional[typing.Sequence[spec.Address]] = None,
    provider: spec.ProviderSpec = None,
    normalize: bool = True,
    usd: bool = False,
    exclude_fei: bool = True,
) -> typing.Union[
    typing.Mapping[spec.Address, list[int]],
    typing.Mapping[spec.Address, list[float]],
]:

    coroutines = [
        async_get_tokens_balances(
            tokens=tokens,
            provider=provider,
            normalize=normalize,
            block=block,
            usd=usd,
            exclude_fei=exclude_fei,
        )
        for block in blocks
    ]
    block_token_balances = await async_utils.gather_coroutines(*coroutines)

    if normalize:
        int_type = list[dict[spec.Address, int]]
        int_result = typing.cast(int_type, block_token_balances)
        return nested_utils.list_of_dicts_to_dict_of_lists(int_result)
    else:
        float_type = list[dict[spec.Address, float]]
        float_result = typing.cast(float_type, block_token_balances)
        return nested_utils.list_of_dicts_to_dict_of_lists(float_result)




# see https://docs.soliditylang.org/en/latest/abi-spec.html

from __future__ import annotations

import typing
from typing_extensions import TypedDict

if typing.TYPE_CHECKING:
    from typing_extensions import NotRequired


class FunctionABI(TypedDict):
    type: typing.Literal['function', 'constructor', 'receive', 'fallback']
    name: NotRequired[str]
    inputs: NotRequired[list['ABIFunctionArg']]
    outputs: NotRequired[list['ABIFunctionArg']]
    stateMutability: typing.Literal['pure', 'view', 'nonpayable', 'payable']


class ABIFunctionArg(TypedDict):
    name: str
    type: 'ABIDatumType'
    components: 'ABITupleComponents'


class EventABI(TypedDict):
    type: typing.Literal['event']
    name: str
    inputs: list['ABIEventArg']


class ABIEventArg(TypedDict):
    name: str
    type: 'ABIDatumType'
    components: 'ABITupleComponents'
    indexed: bool
    anonymous: NotRequired[bool]


class ErrorABI(TypedDict):
    type: typing.Literal['error']
    name: str
    inputs: ABIFunctionArg


ContractABI = typing.List[typing.Union[FunctionABI, EventABI, ErrorABI]]


ABITupleComponents = typing.List['ABITupleComponent']

ABITupleComponent = dict

# # class ABITupleComponent(TypedDict):
# #     name: str
# #     type: ABIDatumType
# #     components: NotRequired['ABITupleComponents']


ABIDatumType = str
ABIDatatypeStr = str
FunctionSelector = str
FunctionSignature = str


class DecodedCallData(TypedDict):
    function_abi: FunctionABI
    function_selector: str
    parameters: list[typing.Any]
    named_parameters: typing.Optional[dict[str, typing.Any]]




import typing
from typing_extensions import TypedDict, Literal

from . import network_types


RpcSingularRequest = typing.Dict[str, typing.Any]
RpcPluralRequest = typing.List[RpcSingularRequest]
RpcRequest = typing.Union[RpcSingularRequest, RpcPluralRequest]


class RpcSingularResponseRaw(TypedDict):
    id: str
    jsonrpc: str
    result: typing.Any


RpcPluralResponseRaw = typing.List[RpcSingularResponseRaw]
RpcResponseRaw = typing.Union[RpcSingularResponseRaw, RpcPluralResponseRaw]

RpcSingularResponse = typing.Any
RpcPluralResponse = typing.List[RpcSingularResponse]
RpcResponse = typing.Union[RpcPluralResponse, RpcSingularResponse]

RpcConstructor = typing.Callable[..., RpcSingularRequest]
RpcDigestor = typing.Callable[..., RpcResponse]


#
# # provider
#

ProviderName = str


class PartialProvider(TypedDict, total=False):
    url: str
    name: typing.Optional[ProviderName]
    network: typing.Optional[network_types.NetworkName]
    protocol: Literal['http', 'wss', 'ipc']
    session_kwargs: typing.Optional[dict]
    chunk_size: typing.Optional[int]


class Provider(TypedDict, total=True):
    url: str
    name: typing.Optional[ProviderName]
    network: typing.Optional[network_types.NetworkName]
    protocol: Literal['http', 'wss', 'ipc']
    session_kwargs: typing.Optional[dict]
    chunk_size: typing.Optional[int]


provider_keys = [
    'url',
    'name',
    'network',
    'protocol',
    'session_kwargs',
    'chunk_size',
]

ProviderShortcut = str
ProviderSpec = typing.Union[ProviderShortcut, PartialProvider, Provider, None]
ProviderKey = typing.Tuple[
    int, str, typing.Tuple[typing.Tuple[typing.Any, typing.Any], ...]
]




from ctc import evm
from ctc import spec

from .. import rpc_provider
from . import rpc_batch_utils


def batch_eth_accounts(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_accounts', **kwargs)


async def async_batch_eth_accounts(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_accounts', **kwargs)


def batch_eth_block_number(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_block_number', **kwargs)


async def async_batch_eth_block_number(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_block_number', **kwargs
    )


def batch_eth_call(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_call', **kwargs)


async def async_batch_eth_call(
    function_abi=None,
    function_name=None,
    function_selector=None,
    provider=None,
    to_address=None,
    to_addresses=None,
    **kwargs,
) -> spec.RpcPluralResponse:

    if function_abi is None:

        if to_address is not None:
            contract_address = to_address
        elif to_addresses is not None:
            contract_address = to_addresses[0]
        else:
            raise Exception(
                'must specify contract_address or contract_addresses'
            )

        provider = rpc_provider.get_provider(provider)
        network = provider['network']
        if network is None:
            raise Exception('could not determine network')

        function_abi = await evm.async_get_function_abi(
            contract_address=contract_address,
            function_name=function_name,
            function_selector=function_selector,
            network=network,
        )

    return await rpc_batch_utils.async_batch_execute(
        'eth_call',
        function_abi=function_abi,
        provider=provider,
        to_address=to_address,
        to_addresses=to_addresses,
        **kwargs,
    )


def batch_eth_coinbase(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_coinbase', **kwargs)


async def async_batch_eth_coinbase(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_coinbase', **kwargs)


def batch_eth_compile_lll(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_compile_lll', **kwargs)


async def async_batch_eth_compile_lll(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_compile_lll', **kwargs
    )


def batch_eth_compile_serpent(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_compile_serpent', **kwargs)


async def async_batch_eth_compile_serpent(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_compile_serpent', **kwargs
    )


def batch_eth_compile_solidity(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_compile_solidity', **kwargs)


async def async_batch_eth_compile_solidity(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_compile_solidity', **kwargs
    )


def batch_eth_estimate_gas(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_estimate_gas', **kwargs)


async def async_batch_eth_estimate_gas(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_estimate_gas', **kwargs
    )


def batch_eth_gas_price(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_gas_price', **kwargs)


async def async_batch_eth_gas_price(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_gas_price', **kwargs)


def batch_eth_get_balance(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_balance', **kwargs)


async def async_batch_eth_get_balance(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_balance', **kwargs
    )


def batch_eth_get_block_by_hash(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_block_by_hash', **kwargs)


async def async_batch_eth_get_block_by_hash(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_block_by_hash', **kwargs
    )


def batch_eth_get_block_by_number(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_block_by_number', **kwargs)


async def async_batch_eth_get_block_by_number(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_block_by_number', **kwargs
    )


def batch_eth_get_block_transaction_count_by_hash(
    **kwargs,
) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_block_transaction_count_by_hash', **kwargs
    )


async def async_batch_eth_get_block_transaction_count_by_hash(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_block_transaction_count_by_hash', **kwargs
    )


def batch_eth_get_block_transaction_count_by_number(
    **kwargs,
) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_block_transaction_count_by_number', **kwargs
    )


async def async_batch_eth_get_block_transaction_count_by_number(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_block_transaction_count_by_number', **kwargs
    )


def batch_eth_get_code(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_code', **kwargs)


async def async_batch_eth_get_code(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_get_code', **kwargs)


def batch_eth_get_compilers(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_compilers', **kwargs)


async def async_batch_eth_get_compilers(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_compilers', **kwargs
    )


def batch_eth_get_filter_changes(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_filter_changes', **kwargs)


async def async_batch_eth_get_filter_changes(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_filter_changes', **kwargs
    )


def batch_eth_get_filter_logs(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_filter_logs', **kwargs)


async def async_batch_eth_get_filter_logs(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_filter_logs', **kwargs
    )


def batch_eth_get_logs(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_logs', **kwargs)


async def async_batch_eth_get_logs(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_get_logs', **kwargs)


def batch_eth_get_storage_at(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_storage_at', **kwargs)


async def async_batch_eth_get_storage_at(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_storage_at', **kwargs
    )


def batch_eth_get_transaction_by_block_hash_and_index(
    **kwargs,
) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_transaction_by_block_hash_and_index', **kwargs
    )


async def async_batch_eth_get_transaction_by_block_hash_and_index(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_transaction_by_block_hash_and_index', **kwargs
    )


def batch_eth_get_transaction_by_block_number_and_index(
    **kwargs,
) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_transaction_by_block_number_and_index', **kwargs
    )


async def async_batch_eth_get_transaction_by_block_number_and_index(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_transaction_by_block_number_and_index', **kwargs
    )


def batch_eth_get_transaction_by_hash(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_transaction_by_hash', **kwargs
    )


async def async_batch_eth_get_transaction_by_hash(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_transaction_by_hash', **kwargs
    )


def batch_eth_get_transaction_count(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_transaction_count', **kwargs)


async def async_batch_eth_get_transaction_count(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_transaction_count', **kwargs
    )


def batch_eth_get_transaction_receipt(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_transaction_receipt', **kwargs
    )


async def async_batch_eth_get_transaction_receipt(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_transaction_receipt', **kwargs
    )


def batch_eth_get_uncle_by_block_hash_and_index(
    **kwargs,
) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_uncle_by_block_hash_and_index', **kwargs
    )


async def async_batch_eth_get_uncle_by_block_hash_and_index(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_uncle_by_block_hash_and_index', **kwargs
    )


def batch_eth_get_uncle_by_block_number_and_index(
    **kwargs,
) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_uncle_by_block_number_and_index', **kwargs
    )


async def async_batch_eth_get_uncle_by_block_number_and_index(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_uncle_by_block_number_and_index', **kwargs
    )


def batch_eth_get_uncle_count_by_block_hash(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_uncle_count_by_block_hash', **kwargs
    )


async def async_batch_eth_get_uncle_count_by_block_hash(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_uncle_count_by_block_hash', **kwargs
    )


def batch_eth_get_uncle_count_by_block_number(
    **kwargs,
) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_get_uncle_count_by_block_number', **kwargs
    )


async def async_batch_eth_get_uncle_count_by_block_number(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_get_uncle_count_by_block_number', **kwargs
    )


def batch_eth_get_work(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_get_work', **kwargs)


async def async_batch_eth_get_work(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_get_work', **kwargs)


def batch_eth_hashrate(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_hashrate', **kwargs)


async def async_batch_eth_hashrate(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_hashrate', **kwargs)


def batch_eth_mining(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_mining', **kwargs)


async def async_batch_eth_mining(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_mining', **kwargs)


def batch_eth_new_block_filter(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_new_block_filter', **kwargs)


async def async_batch_eth_new_block_filter(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_new_block_filter', **kwargs
    )


def batch_eth_new_filter(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_new_filter', **kwargs)


async def async_batch_eth_new_filter(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_new_filter', **kwargs)


def batch_eth_new_pending_transaction_filter(
    **kwargs,
) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute(
        'eth_new_pending_transaction_filter', **kwargs
    )


async def async_batch_eth_new_pending_transaction_filter(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_new_pending_transaction_filter', **kwargs
    )


def batch_eth_protocol_version(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_protocol_version', **kwargs)


async def async_batch_eth_protocol_version(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_protocol_version', **kwargs
    )


def batch_eth_send_raw_transaction(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_send_raw_transaction', **kwargs)


async def async_batch_eth_send_raw_transaction(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_send_raw_transaction', **kwargs
    )


def batch_eth_send_transaction(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_send_transaction', **kwargs)


async def async_batch_eth_send_transaction(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_send_transaction', **kwargs
    )


def batch_eth_sign(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_sign', **kwargs)


async def async_batch_eth_sign(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_sign', **kwargs)


def batch_eth_sign_transaction(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_sign_transaction', **kwargs)


async def async_batch_eth_sign_transaction(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_sign_transaction', **kwargs
    )


def batch_eth_submit_hashrate(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_submit_hashrate', **kwargs)


async def async_batch_eth_submit_hashrate(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_submit_hashrate', **kwargs
    )


def batch_eth_submit_work(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_submit_work', **kwargs)


async def async_batch_eth_submit_work(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_submit_work', **kwargs
    )


def batch_eth_syncing(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_syncing', **kwargs)


async def async_batch_eth_syncing(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('eth_syncing', **kwargs)


def batch_eth_uninstall_filter(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('eth_uninstall_filter', **kwargs)


async def async_batch_eth_uninstall_filter(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'eth_uninstall_filter', **kwargs
    )


def batch_net_listening(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('net_listening', **kwargs)


async def async_batch_net_listening(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('net_listening', **kwargs)


def batch_net_peer_count(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('net_peer_count', **kwargs)


async def async_batch_net_peer_count(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('net_peer_count', **kwargs)


def batch_net_version(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('net_version', **kwargs)


async def async_batch_net_version(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('net_version', **kwargs)


def batch_shh_add_to_group(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_add_to_group', **kwargs)


async def async_batch_shh_add_to_group(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'shh_add_to_group', **kwargs
    )


def batch_shh_get_filter_changes(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_get_filter_changes', **kwargs)


async def async_batch_shh_get_filter_changes(
    **kwargs,
) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'shh_get_filter_changes', **kwargs
    )


def batch_shh_get_messages(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_get_messages', **kwargs)


async def async_batch_shh_get_messages(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'shh_get_messages', **kwargs
    )


def batch_shh_has_identity(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_has_identity', **kwargs)


async def async_batch_shh_has_identity(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'shh_has_identity', **kwargs
    )


def batch_shh_new_filter(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_new_filter', **kwargs)


async def async_batch_shh_new_filter(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('shh_new_filter', **kwargs)


def batch_shh_new_group(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_new_group', **kwargs)


async def async_batch_shh_new_group(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('shh_new_group', **kwargs)


def batch_shh_new_identity(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_new_identity', **kwargs)


async def async_batch_shh_new_identity(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'shh_new_identity', **kwargs
    )


def batch_shh_post(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_post', **kwargs)


async def async_batch_shh_post(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('shh_post', **kwargs)


def batch_shh_uninstall_filter(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_uninstall_filter', **kwargs)


async def async_batch_shh_uninstall_filter(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'shh_uninstall_filter', **kwargs
    )


def batch_shh_version(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('shh_version', **kwargs)


async def async_batch_shh_version(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('shh_version', **kwargs)


def batch_web3_client_version(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('web3_client_version', **kwargs)


async def async_batch_web3_client_version(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute(
        'web3_client_version', **kwargs
    )


def batch_web3_sha3(**kwargs) -> spec.RpcPluralResponse:
    return rpc_batch_utils.batch_execute('web3_sha3', **kwargs)


async def async_batch_web3_sha3(**kwargs) -> spec.RpcPluralResponse:
    return await rpc_batch_utils.async_batch_execute('web3_sha3', **kwargs)




import subprocess


def get_directory_nbytes(dirpath: str) -> int:
    cmd = 'du -sb ' + dirpath
    cmd_pieces = cmd.split(' ')
    output_bytes = subprocess.check_output(cmd_pieces)
    output = output_bytes.split(b'\t')[0].decode()
    return int(output)


def get_directory_nbytes_human(dirpath: str) -> str:
    cmd = 'du -h ' + dirpath
    cmd_pieces = cmd.split(' ')
    output = subprocess.check_output(cmd_pieces)
    return output.split(b'\t')[0].decode()




"""

## Config Keys
- version: version of config schema
- data_dir: root directory for storing ctc data
- default_network: default network to use when none specified
- providers: default provider for each network
- networks: custom user-defined networks


## TODO
- add additional config settings
    - rpc: set batching and other parameters for each rpc method
    - sql: configuration for sql database backend
"""

import typing
from typing_extensions import TypedDict

from . import network_types
from . import rpc_types


class ConfigNetworkDefaults(TypedDict):
    default_network: network_types.NetworkName
    default_providers: typing.Dict[
        network_types.NetworkName, rpc_types.ProviderName
    ]


class PartialConfigSpec(TypedDict, total=False):
    config_spec_version: str
    data_dir: str
    providers: typing.Dict[rpc_types.ProviderName, rpc_types.Provider]
    networks: typing.Dict[
        network_types.NetworkName, network_types.NetworkMetadata
    ]
    network_defaults: ConfigNetworkDefaults


class ConfigSpec(TypedDict):
    config_spec_version: str
    data_dir: str
    providers: typing.Dict[rpc_types.ProviderName, rpc_types.Provider]
    networks: typing.Dict[
        network_types.NetworkName, network_types.NetworkMetadata
    ]
    network_defaults: ConfigNetworkDefaults




import math
import os
import typing

import toolstr

from ctc import evm
from ctc import binary
from ctc import spec

from . import filesystem_events
from ... import abi_utils


async def async_rechunk_events(
    contract_address: spec.Address,
    chunk_target_bytes: int,
    event_name: typing.Optional[str] = None,
    event_hash: typing.Optional[str] = None,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    chunk_size_tolerance: float = 0.1,
    verbose: bool = True,
    dry: bool = False,
    network: spec.NetworkReference = 'mainnet',
):
    import numpy as np

    # get event hash
    if event_hash is None:
        if event_name is None:
            raise Exception('must specify event_name or event_name')
        event_abi = await abi_utils.async_get_event_abi(
            contract_address=contract_address, event_name=event_name
        )
        event_name = binary.get_event_hash(event_abi)

    # make sure no leftover files from previous rechunks
    event_list = filesystem_events.list_events(
        contract_address=contract_address,
        event_hash=event_hash,
        network=network,
    )
    if event_list is None:
        if verbose:
            print('skipping rechunk, no event data found')
        return
    for path in event_list['paths']:
        if path.endswith('__OLD'):
            event_dir = filesystem_events.get_events_event_dir(
                contract_address=contract_address,
                event_hash=event_hash,
                network=network,
            )
            message = (
                'there exists <filename>__OLD files from an incomplete'
                ' rechunking, remove these files to continue in directory '
            ) + str(event_dir)
            raise Exception(message)

    # load events
    if start_block is not None:
        start_block = await evm.async_block_number_to_int(start_block)
    if end_block is not None:
        end_block = await evm.async_block_number_to_int(end_block)

    # get event range
    if start_block is None:
        start_block = min(
            path_start for path_start, path_end in event_list['paths'].values()
        )
    if end_block is None:
        end_block = max(
            path_end for path_start, path_end in event_list['paths'].values()
        )
    start_block = typing.cast(int, start_block)
    end_block = typing.cast(int, end_block)

    # get original files of block range
    original_paths = []
    for path, (path_start, path_end) in event_list['paths'].items():
        if (start_block <= path_start and path_start <= end_block) or (
            start_block <= end_block and end_block <= end_block
        ):
            original_paths.append(path)
    original_paths = sorted(original_paths)

    # eliminate paths that are within chunk size tolerance
    path_sizes = {}
    min_path_size = chunk_target_bytes * (1 - chunk_size_tolerance)
    max_path_size = chunk_target_bytes * (1 + chunk_size_tolerance)
    n_trim_start = 0
    for original_path in original_paths:
        path_bytes = os.path.getsize(original_path)
        path_sizes[original_path] = path_bytes
        if min_path_size <= path_bytes and path_bytes <= max_path_size:
            n_trim_start += 1
        else:
            break
    original_paths = original_paths[n_trim_start:]
    n_trim_end = 0
    for original_path in original_paths[::-1]:
        if original_path not in path_sizes:
            path_bytes = os.path.getsize(original_path)
            path_sizes[original_path] = path_bytes
        if min_path_size <= path_bytes and path_bytes <= max_path_size:
            n_trim_end += 1
        else:
            break
    if n_trim_end > 0:
        original_paths = original_paths[:-n_trim_end]

    if len(original_paths) == 1 and path_sizes[original_path] < max_path_size:
        original_paths = []

    # get block range of target paths
    min_block_start = int(float(1e99))
    max_block_end = -1
    for path in original_paths:
        path_start, path_end = event_list['paths'][path]
        min_block_start = min(path_start, min_block_start)
        max_block_end = max(path_end, max_block_end)

    # compute total bytes across all files
    data_bytes = 0
    for original_path in original_paths:
        if original_path not in path_sizes:
            path_sizes[original_path] = os.path.getsize(original_path)
        data_bytes += path_sizes[original_path]

    # compute number of chunks
    n_chunks = math.ceil(data_bytes / chunk_target_bytes)

    # load events
    if len(original_paths) > 0:
        events = await filesystem_events.async_get_events_from_filesystem(
            contract_address=contract_address,
            event_hash=event_hash,
            start_block=min_block_start,
            end_block=max_block_end,
            network=network,
        )

    # split into chunks
    if len(original_paths) > 0 and len(events) > 0:
        event_chunks = typing.cast(
            typing.Sequence[spec.DataFrame],
            np.array_split(events, n_chunks),
        )
    else:
        event_chunks = []

    # get each chunk's block range
    chunk_ranges = []
    for event_chunk in event_chunks:
        chunk_start_block = event_chunk.index.values[0][0]
        chunk_end_block = event_chunk.index.values[-1][0]
        chunk_ranges.append([chunk_start_block, chunk_end_block])

    # remove gaps between block ranges
    for e, (start_block, end_block) in enumerate(chunk_ranges):
        if e + 1 < len(chunk_ranges):
            chunk_ranges[e][1] = chunk_ranges[e + 1][0] - 1

    # remove gaps at beginning and end of ranges
    if len(chunk_ranges) > 0:
        chunk_ranges[0][0] = min(min_block_start, chunk_ranges[0][0])
        chunk_ranges[-1][1] = max(max_block_end, chunk_ranges[-1][1])

    # print summary
    if verbose or dry:
        event_dir = filesystem_events.get_events_event_dir(
            contract_address=contract_address,
            event_hash=event_hash,
            network=network,
        )
        if dry:
            print('[DRY RUN -- WILL NOT CHANGE FILES]')
        print('event rechunking')
        print('- contract:', contract_address)
        print('- event_hash:', event_hash)
        print('- block range', [start_block, end_block])
        print('- data_bytes:', toolstr.format(data_bytes))
        print('- directory:', event_dir)
        print()
        if len(chunk_ranges) == 0:
            print('current chunks acceptable, no rechunking required')
        else:
            print(
                'rechunking',
                len(original_paths),
                'files into',
                n_chunks,
                'files',
            )

    if dry:
        print('[DRY RUN -- NO FILES MODIFIED]')
        return

    if len(chunk_ranges) == 0:
        return

    # rename original files
    for original_path in original_paths:
        os.rename(original_path, original_path + '__OLD')

    # save chunks
    for event_chunk, chunk_range in zip(event_chunks, chunk_ranges):
        chunk_start_block, chunk_end_block = chunk_range
        path = await filesystem_events.async_save_events_to_filesystem(
            events=event_chunk,
            contract_address=contract_address,
            event_hash=event_hash,
            start_block=chunk_start_block,
            end_block=chunk_end_block,
            network=network,
        )

    # remove original files
    for original_path in original_paths:
        os.remove(original_path + '__OLD')

    if verbose:
        print()
        print('...done')


async def async_rechunk_all_events(
    chunk_target_bytes: int,
    network: spec.NetworkReference = 'mainnet',
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    chunk_size_tolerance: float = 0.1,
    verbose: bool = True,
    dry: bool = False,
):
    contracts_events = filesystem_events.list_contracts_events(network=network)

    for contract_address, events_data in contracts_events.items():
        for event_hash, event_data in events_data.items():
            await async_rechunk_events(
                contract_address=contract_address,
                event_hash=event_hash,
                start_block=start_block,
                end_block=end_block,
                chunk_target_bytes=chunk_target_bytes,
                chunk_size_tolerance=chunk_size_tolerance,
                verbose=verbose,
                dry=dry,
            )
            if verbose:
                print()
                print()




from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_eth_new_filter(
    address=None,
    topics=None,
    start_block=None,
    end_block=None,
    provider=None,
    decode_response=False,
):
    request = rpc_constructors.construct_eth_new_filter(
        address=address,
        topics=topics,
        start_block=start_block,
        end_block=end_block,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_new_filter(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_new_block_filter(provider=None, decode_response=False):
    request = rpc_constructors.construct_eth_new_block_filter()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_new_block_filter(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_new_pending_transaction_filter(
    provider=None, decode_response=False
):
    request = rpc_constructors.construct_eth_new_pending_transaction_filter()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_new_pending_transaction_filter(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_uninstall_filter(
    filter_id, provider=None, decode_response=False
):
    request = rpc_constructors.construct_eth_uninstall_filter(filter_id=filter_id)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_uninstall_filter(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_get_filter_changes(
    filter_id,
    provider=None,
    decode_response=True,
    snake_case_response=True,
    include_removed=False,
):
    request = rpc_constructors.construct_eth_get_filter_changes(filter_id=filter_id)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_filter_changes(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
        include_removed=include_removed,
    )


async def async_eth_get_filter_logs(
    filter_id,
    provider=None,
    decode_response=True,
    snake_case_response=True,
    include_removed=False,
):
    request = rpc_constructors.construct_eth_get_filter_logs(filter_id=filter_id)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_filter_logs(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
        include_removed=include_removed,
    )


async def async_eth_get_logs(
    address=None,
    topics=None,
    start_block=None,
    end_block=None,
    block_hash=None,
    provider=None,
    decode_response=True,
    snake_case_response=True,
    include_removed=False,
):
    request = rpc_constructors.construct_eth_get_logs(
        address=address,
        topics=topics,
        start_block=start_block,
        end_block=end_block,
        block_hash=block_hash,
    )
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_get_logs(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
        include_removed=include_removed,
    )




from __future__ import annotations

import asyncio
import concurrent.futures
import functools
import typing
import types

from typing_extensions import ParamSpec
import pandas as pd


T = typing.TypeVar('T')
R = typing.TypeVar('R')

#
# # gathering
#


async def gather_coroutines(
    *coroutines: typing.Coroutine[typing.Any, typing.Any, R]
) -> list[R]:
    """gather without erasing type information"""
    return await asyncio.gather(*coroutines)


async def gather_dict(
    d: typing.Mapping[T, typing.Coroutine[typing.Any, typing.Any, R]]
) -> dict[T, R]:

    tasks = list(d.values())
    results = await asyncio.gather(*tasks)
    return dict(zip(d.keys(), results))


# typing_extensions does not support ParamSpec properly


# P = ParamSpec('P')


# def sync_wrap(f: typing.Callable[P, R]) -> types.FunctionType:
#     """decorator to create synchronous versions of asynchronous functions

#     - see if anyio has the capability for this already

#     - if currently in an eventloop, f should run in that eventloop
#     - if not currently in an eventloop, f should start one
#     """

#     @functools.wraps(f)
#     def new_f(*args: P.args, **kwargs: P.kwrags) -> R:
#         return asyncio.run(f(*args, **kwargs))

#     return new_f


# async def gather_nested(
#     nested: dict[typing.Any, typing.Any],
# ) -> dict[typing.Any, typing.Any]:

#     # compile tasks
#     tasks = {}
#     for key, value in nested.items():
#         if isinstance(value, dict):
#             tasks[key] = gather_nested(value)
#         elif hasattr(value, '__await__'):
#             tasks[key] = value

#     # await results
#     results = await gather_dict(tasks)

#     # compile output
#     output = {}
#     for key, value in nested.items():
#         if key in results:
#             output[key] = results[key]
#         else:
#             output[key] = value

#     return output


##
## # parallelizing
##

# from typing_extensions import ParamSpec
# ParamSpec = ParamSpec('P')
# T = typing.TypeVar('T')


# async def parallelize(
#    arg: str,
#    values: list[typing.Any],
#    function: typing.Callable[P, T],
#    kwargs: P.kwargs,
# ) -> list[T]:
#    coroutines = []
#    for value in values:
#        coroutine = function(**{arg: value}, **kwargs)
#        coroutines.append(coroutine)
#    return await asyncio.gather(*coroutines)


#
# # csv loading
#


# TODO: many possible solutions to speeding up csv reading + asychronousity
# - shared memory
#     - https://stackoverflow.com/questions/7894791/use-numpy-array-in-shared-memory-for-multiprocessing
#     - https://stackoverflow.com/questions/10721915/shared-memory-objects-in-multiprocessing
#     - https://stackoverflow.com/questions/14124588/shared-memory-in-multiprocessing
# - use seek(n) to have different processes read different parts of same file
# - break into more regularly sized files
# - use hdf instead of csv
# - use database instead of csv
# - things to note:
#    - pandas's read_csv is faster than other python csv libs
#    - the bottleneck is cpu after data is loaded, NOT diskio
#    - if crossing process boundary, data transfers will become an issue


async def async_read_csv(path, mode=None, **kwargs):

    if mode is None:
        mode == 'processes'

    if mode == 'processes':
        return await async_read_csv_processes(path, **kwargs)
    if mode == 'threads':
        return await async_read_csv_threads(path, **kwargs)
    elif mode == 'dask':
        return await async_read_csv_dask(path, **kwargs)
    elif mode == 'aiofiles':
        return await async_read_csv_aiofiles(path, **kwargs)
    elif mode == 'anyio':
        return await async_read_csv_anyio(path, **kwargs)
    else:
        raise Exception('unknown mode: ' + str(mode))


async def async_read_csv_processes(path, pool=None):
    loop = asyncio.get_running_loop()

    if pool is None:
        with concurrent.futures.ProcessPoolExecutor() as pool:
            return await loop.run_in_executor(
                pool,
                functools.partial(pd.read_csv, path),
            )
    else:
        return await loop.run_in_executor(
            pool,
            functools.partial(pd.read_csv, path),
        )


async def async_read_csv_threads(path, pool=None):
    loop = asyncio.get_running_loop()

    if pool is None:
        with concurrent.futures.ThreadPoolExecutor() as pool:
            return await loop.run_in_executor(
                pool,
                functools.partial(pd.read_csv, path),
            )
    else:
        return await loop.run_in_executor(
            pool,
            functools.partial(pd.read_csv, path),
        )


async def async_read_csv_dask(path):
    import dask.dataframe as dd  # type: ignore
    import dask.distributed  # type: ignore

    client = await dask.distributed.Client(asynchronous=True)
    g = dd.read_csv(path, dtype='object')
    return await client.compute(g)


async def async_read_csv_aiofiles(path):
    import io
    import aiofiles

    async with aiofiles.open(path, 'r') as f:
        contents = await f.read()
        string_io = io.StringIO(contents)
        # return string_io
        return pd.read_csv(string_io)


async def async_read_csv_anyio(path):
    import anyio.to_process

    return await anyio.to_process.run_sync(pd.read_csv, path)




from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


async def async_web3_client_version(provider=None):
    request = rpc_constructors.construct_web3_client_version()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_web3_client_version(response=response)


async def async_web3_sha3(data, provider=None):
    request = rpc_constructors.construct_web3_sha3(data)
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_web3_sha3(response=response)


async def async_net_version(provider=None):
    request = rpc_constructors.construct_net_version()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_net_version(response=response)


async def async_net_peer_count(provider=None):
    request = rpc_constructors.construct_net_peer_count()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_net_peer_count(response=response)


async def async_net_listening(provider=None):
    request = rpc_constructors.construct_net_listening()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_net_listening(response=response)


async def async_eth_protocol_version(provider=None, decode_response=True):
    request = rpc_constructors.construct_eth_protocol_version()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_protocol_version(
        response=response,
        decode_response=decode_response,
    )


async def async_eth_syncing(provider=None, snake_case_response=True):
    request = rpc_constructors.construct_eth_syncing()
    response = await rpc_request.async_send(request, provider=provider)
    return rpc_digestors.digest_eth_syncing(
        response=response,
        snake_case_response=snake_case_response,
    )




import os
import subprocess

import ctc.config


def get_command_spec():
    return {
        'f': edit_config_command,
    }


def edit_config_command():
    editor = os.environ.get('EDITOR')
    if editor is None or editor == '':
        raise Exception('set $EDITOR env var')

    config_path = ctc.config.get_config_path()

    print('editing config in editor')
    print('- editor:', editor)
    print('- config_path:', config_path)
    print()
    subprocess.call([editor, config_path])
    print('done editing config')




import copy
import os
import typing

import ctc.config
from ctc import spec


def get_provider(provider: spec.ProviderSpec = None) -> spec.Provider:

    if provider is None:

        # case: return default provider
        return ctc.config.get_default_provider()

    elif isinstance(provider, str):

        # case: provider specified as url
        if provider.startswith('http'):
            return {
                'name': None,
                'network': None,
                'protocol': 'http',
                'url': provider,
                'session_kwargs': {},
                'chunk_size': None,
            }

        # case: provider specified as name in config
        elif ctc.config.has_provider(name=provider):
            return ctc.config.get_provider(name=provider)

        else:
            raise Exception('unknown provider format: ' + str(provider))

    elif isinstance(provider, dict):

        # case: partial provider
        if set(provider.keys()) == spec.provider_keys:
            return typing.cast(spec.Provider, provider)

        else:
            selection_keys = ['name', 'network', 'protocol']
            if any(provider.get(key) is not None for key in selection_keys):
                base_provider = ctc.config.get_provider(
                    name=provider.get('name'),
                    network=provider.get('network'),
                    protocol=provider.get('protocol'),
                )
            else:
                base_provider = ctc.config.get_default_provider()

            non_none_keys = typing.cast(
                spec.PartialProvider,
                {k: v for k, v in provider.items() if v is not None}
            )
            full_provider = copy.copy(base_provider)
            full_provider.update(non_none_keys)

            return full_provider

    else:
        raise Exception('unknown provider type: ' + str(type(provider)))


def get_provider_key(provider: spec.Provider) -> spec.ProviderKey:
    """return a unique identifier for provider within this process"""

    session_kwargs = provider.get('session_kwargs')
    if session_kwargs is None:
        session_kwargs = {}

    return (os.getpid(), provider['url'], tuple(session_kwargs.items()))


def add_provider_parameters(
    provider: spec.ProviderSpec,
    parameters: spec.PartialProvider,
) -> spec.Provider:
    # TODO: decide whether parameters with value=None should be included
    provider = get_provider(provider)
    provider = copy.copy(provider)
    provider.update(parameters)
    return provider




from __future__ import annotations

import asyncio
import typing

from ctc import rpc
from ctc import spec


async def async_get_eth_total_supply():
    import aiohttp

    url = 'http://api.etherscan.io/api?module=stats&action=ethsupply'
    session = aiohttp.ClientSession()
    async with session.get(url) as response:
        if response.status != 200:
            raise Exception('could not get result')
        result = await response.json()
        return int(result['result'])


#
# # balances
#


@typing.overload
async def async_get_eth_balance(
    address: spec.Address,
    normalize: typing.Literal[False],
    provider: typing.Optional[spec.ProviderSpec] = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> int:
    ...


@typing.overload
async def async_get_eth_balance(
    address: spec.Address,
    normalize: bool = True,
    provider: typing.Optional[spec.ProviderSpec] = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> float:
    ...


async def async_get_eth_balance(
    address: spec.Address,
    normalize: bool = True,
    provider: typing.Optional[spec.ProviderSpec] = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> typing.Union[int, float]:

    balance = await rpc.async_eth_get_balance(
        address=address,
        provider=provider,
        block_number=block,
    )

    if normalize:
        balance /= 1e18

    return balance


@typing.overload
async def async_get_eth_balance_by_block(
    address: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: typing.Literal[False],
    provider: typing.Optional[spec.ProviderSpec] = None,
) -> list[int]:
    ...


@typing.overload
async def async_get_eth_balance_by_block(
    address: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: typing.Literal[True] = True,
    provider: typing.Optional[spec.ProviderSpec] = None,
) -> list[float]:
    ...


async def async_get_eth_balance_by_block(
    address: spec.Address,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
    provider: typing.Optional[spec.ProviderSpec] = None,
) -> typing.Union[list[int], list[float]]:

    coroutines = []
    for block in blocks:
        coroutine = async_get_eth_balance(
            address=address, provider=provider, block=block, normalize=normalize
        )
        coroutines.append(coroutine)

    return await asyncio.gather(*coroutines)


@typing.overload
async def async_get_eth_balance_of_addresses(
    addresses: spec.Address,
    block: typing.Sequence[spec.BlockNumberReference],
    normalize: typing.Literal[False],
    provider: typing.Optional[spec.ProviderSpec] = None,
) -> list[int]:
    ...


@typing.overload
async def async_get_eth_balance_of_addresses(
    addresses: spec.Address,
    block: typing.Sequence[spec.BlockNumberReference],
    normalize: typing.Literal[True] = True,
    provider: typing.Optional[spec.ProviderSpec] = None,
) -> list[float]:
    ...


async def async_get_eth_balance_of_addresses(
    addresses: typing.Sequence[spec.Address],
    normalize: bool = True,
    provider: typing.Optional[spec.ProviderSpec] = None,
    block: typing.Optional[spec.BlockNumberReference] = None,
) -> typing.Union[list[int], list[float]]:

    balances = await rpc.async_batch_eth_get_balance(
        addresses=addresses,
        provider=provider,
        block_number=block,
    )

    if normalize:
        balances = [balance / 1e18 for balance in balances]

    return balances




import typing
from typing_extensions import Literal


BinaryData = typing.Union[bytes, str]
BinaryInteger = typing.Union[int, BinaryData]
BlockSpec = typing.Union[
    int, BinaryData, Literal['latest', 'earliest', 'pending']
]

BinaryFormat = Literal['binary', 'prefix_hex', 'raw_hex', 'integer']

IntegerData = int

PrefixHexData = str
RawHexData = str
HexData = typing.Union[PrefixHexData, RawHexData]

Data = typing.Union[IntegerData, BinaryData, HexData]




import typing

from ctc import spec

import toolconfig


def write_config_file(
    config_data: spec.ConfigSpec,
    path: str,
    overwrite: bool = False,
    style: typing.Optional[str] = None
) -> None:

    toolconfig.write_config_file(
        config_data=typing.cast(
            typing.MutableMapping[str, typing.Any],
            config_data,
        ),
        path=path,
        overwrite=overwrite,
        style=style,
    )




import typing
from typing_extensions import TypedDict, Literal


Address = str
ContractAddress = str
ERC20Address = ContractAddress
ERC20Symbol = str
ERC20Reference = typing.Union[ERC20Address, ERC20Symbol]


class AddressRequiredMetadata(TypedDict):
    name: str  # human readable name
    address: Address  # 20 byte address data


class AddressMetadata(AddressRequiredMetadata, total=False):
    first_block: typing.Optional[int]  # the block this address started existing
    contract_name: typing.Optional[str]  # name as appears in verified contract
    label: str  # protocol name, project name, etc


Oracletype = Literal['feed', 'amm']


class OracleFeedMetadata(AddressMetadata, total=True):
    numerator: typing.Optional[str]
    denominator: typing.Optional[str]
    deviation: float  # percentage
    heartbeat: int  # number of seconds
    decimals: int


class ERC20Metadata(AddressMetadata):
    symbol: str
    decimals: int




from ctc import spec
from ctc import binary
from .. import rpc_format
from .. import rpc_spec


def digest_eth_block_number(
    response: spec.RpcSingularResponse, decode_response: bool = True
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_get_block_by_hash(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        quantities = rpc_spec.rpc_block_quantities
        response = rpc_format.decode_response(response, quantities=quantities)

    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)

    return response


def digest_eth_get_block_by_number(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:

    if decode_response:
        quantities = rpc_spec.rpc_block_quantities
        response = rpc_format.decode_response(response, quantities=quantities)

    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)

    include_full_transactions = len(
        response['transactions']
    ) > 0 and isinstance(response['transactions'][0], dict)
    if include_full_transactions and (decode_response or snake_case_response):

        transaction_quantities = rpc_spec.rpc_transaction_quantities
        new_transactions = []
        for transaction in response['transactions']:

            new_transaction = transaction

            if decode_response:
                new_transaction = rpc_format.decode_response(
                    new_transaction, quantities=transaction_quantities
                )

            if snake_case_response:
                new_transaction = rpc_format.keys_to_snake_case(new_transaction)
            new_transactions.append(new_transaction)

        response['transactions'] = new_transactions

    return response


def digest_eth_get_uncle_count_by_block_hash(
    response: spec.RpcSingularResponse, decode_response: bool = True
):
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_get_uncle_count_by_block_number(
    response: spec.RpcSingularResponse, decode_response: bool = True
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_get_uncle_by_block_hash_and_index(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        quantities = rpc_spec.rpc_block_quantities
        response = rpc_format.decode_response(response, quantities=quantities)

    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)

    return response


def digest_eth_get_uncle_by_block_number_and_index(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        quantities = rpc_spec.rpc_block_quantities
        response = rpc_format.decode_response(response, quantities=quantities)

    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)

    return response




import pytest

from ctc import rpc


@pytest.mark.asyncio
async def test_eth_get_transaction_count():
    result = await rpc.async_eth_get_transaction_count(
        '0x94b0a3d511b6ecdb17ebf877278ab030acb0a878'
    )
    assert result == 1


@pytest.mark.asyncio
async def test_eth_get_transaction_by_block_hash_and_index():
    result = await rpc.async_eth_get_transaction_by_block_hash_and_index(
        block_hash='0xb2a8b39935a5eb4b7c9b0117bca06c8d2c0629e0937d20e62c44aace6f05bda3',
        transaction_index=88,
    )
    assert (
        result['block_hash']
        == '0xb2a8b39935a5eb4b7c9b0117bca06c8d2c0629e0937d20e62c44aace6f05bda3'
    )


@pytest.mark.asyncio
async def test_eth_get_transaction_by_block_number_and_index():
    result = await rpc.async_eth_get_transaction_by_block_number_and_index(
        block_number=12345678,
        transaction_index=88,
    )
    assert (
        result['block_hash']
        == '0xb2a8b39935a5eb4b7c9b0117bca06c8d2c0629e0937d20e62c44aace6f05bda3'
    )


@pytest.mark.asyncio
async def test_eth_get_transaction_by_hash():
    result = await rpc.async_eth_get_transaction_by_hash(
        transaction_hash='0xd764f7b1bfc34acfa052453ee1baf2d016d0c8c4d963a73db5e30780a6f92e7b'
    )
    assert (
        result['block_hash']
        == '0xb2a8b39935a5eb4b7c9b0117bca06c8d2c0629e0937d20e62c44aace6f05bda3'
    )


@pytest.mark.asyncio
async def test_eth_get_transaction_receipt():
    result = await rpc.async_eth_get_transaction_receipt(
        '0xd764f7b1bfc34acfa052453ee1baf2d016d0c8c4d963a73db5e30780a6f92e7b'
    )
    assert (
        result['block_hash']
        == '0xb2a8b39935a5eb4b7c9b0117bca06c8d2c0629e0937d20e62c44aace6f05bda3'
    )


@pytest.mark.asyncio
async def test_get_block_transaction_count_by_hash():
    result = await rpc.async_eth_get_block_transaction_count_by_hash(
        block_hash='0xb2a8b39935a5eb4b7c9b0117bca06c8d2c0629e0937d20e62c44aace6f05bda3'
    )
    assert result == 174


@pytest.mark.asyncio
async def test_eth_get_block_transaction_count_by_number():
    result = await rpc.async_eth_get_block_transaction_count_by_number(
        block_number=12345678
    )
    assert result == 174




from __future__ import annotations

import typing
from typing_extensions import Literal

from ctc import spec
from . import formats


@typing.overload
def keccak(
    data: spec.BinaryInteger,
    output_format: Literal['integer'],
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> int:
    ...


@typing.overload
def keccak(
    data: spec.BinaryInteger,
    output_format: Literal['binary'],
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> bytes:
    ...


@typing.overload
def keccak(
    data: spec.BinaryInteger,
    output_format: Literal['prefix_hex', 'raw_hex', None],
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> str:
    ...


def keccak(
    data: spec.BinaryInteger,
    output_format: typing.Optional[spec.BinaryFormat] = 'prefix_hex',
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> spec.BinaryInteger:
    """return keccack-256 hash of hex or binary data"""

    # determine library
    if library is None:
        try:
            import sha3  # type: ignore

            library = 'pysha3'

        except ImportError:
            library = 'pycryptodome'

    # convert data to binary
    data = formats.convert(data, 'binary')

    if library == 'pysha3':
        import sha3  # type: ignore

        binary = sha3.keccak_256(data).digest()
    elif library == 'pycryptodome':
        from Crypto.Hash import keccak as f_keccak

        binary = f_keccak.new(digest_bits=256, data=data).digest()
    else:
        raise Exception(
            'must choose valid library, either \'pysha3\' or \'pycryptodome\''
        )

    return formats.convert(binary, output_format)


@typing.overload
def keccak_text(
    text: str,
    output_format: Literal['integer'],
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> int:
    ...


@typing.overload
def keccak_text(
    text: str,
    output_format: Literal['binary'],
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> bytes:
    ...


@typing.overload
def keccak_text(
    text: str,
    output_format: Literal['prefix_hex', 'raw_hex'] = 'prefix_hex',
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> str:
    ...


def keccak_text(
    text: str,
    output_format: spec.BinaryFormat = 'prefix_hex',
    library: typing.Optional[typing.Literal['pysha3', 'pycryptodome']] = None,
) -> spec.BinaryInteger:
    """return keccack-256 hash of text"""
    return keccak(text.encode(), output_format=output_format, library=library)




import math

from ctc import evm
from ctc import rpc
from ctc import directory

from . import irm_metadata
from . import pool_metadata
from . import pool_state
from . import token_metadata
from . import token_state


async def async_get_pool_summary(comptroller, block='latest'):

    # get block number
    if block == 'latest':
        block = await rpc.async_eth_block_number()

    # get pool data
    ctokens = await pool_metadata.async_get_pool_ctokens(
        comptroller, block=block
    )
    underlyings = await pool_metadata.async_get_pool_underlying_tokens(
        ctokens=ctokens, block=block
    )
    token_names = []
    for underlying in underlyings.values():
        if underlying == '0x0000000000000000000000000000000000000000':
            token_name = 'ETH'
        elif directory.has_erc20_metadata(address=underlying):
            token_name = directory.get_erc20_symbol(address=underlying)
        else:
            token_name = await evm.async_get_erc20_symbol(underlying)
        token_names.append(token_name)

    # get pricing data
    token_prices = await pool_state.async_get_pool_prices(
        ctokens=ctokens, comptroller=comptroller, block=block
    )

    tokens_data = {}
    blocks_per_year = None
    for ctoken, underlying, token_name in zip(
        ctokens, underlyings, token_names
    ):

        # get blocks per year
        if blocks_per_year is None:
            interest_rate_model = (
                await token_metadata.async_get_ctoken_irm(
                    ctoken=ctoken,
                    block=block,
                )
            )
            blocks_per_year = await irm_metadata.async_get_irm_blocks_per_year(
                interest_rate_model=interest_rate_model,
                block=block,
            )

        # get ctoken stats
        borrowed = await token_state.async_get_total_borrowed(
            ctoken=ctoken, block=block
        )
        borrowed /= 1e18
        liquidity = await token_state.async_get_total_liquidity(
            ctoken=ctoken, block=block
        )
        liquidity /= 1e18
        supply_apy = await token_state.async_get_supply_apy(
            ctoken=ctoken,
            block=block,
            blocks_per_year=blocks_per_year,
        )
        borrow_apy = await token_state.async_get_borrow_apy(
            ctoken=ctoken,
            block=block,
            blocks_per_year=blocks_per_year,
        )

        # compute derived stats
        supplied = borrowed + liquidity
        if math.isclose(supplied, 0):
            utilization = 0
        else:
            utilization = borrowed / supplied

        tokens_data[token_name] = {
            'name': token_name,
            'borrowed': borrowed,
            'borrowed_tvl': borrowed * token_prices[ctoken],
            'supplied': supplied,
            'supplied_tvl': supplied * token_prices[ctoken],
            'liquidity': liquidity,
            'liquidity_tvl': liquidity * token_prices[ctoken],
            'utilization': utilization,
            'borrow_apy': borrow_apy,
            'supply_apy': supply_apy,
        }

    return tokens_data




from __future__ import annotations

import typing

from ctc import spec
from . import erc20_generic


#
# # decimals
#


async def async_get_erc20_decimals(
    token: spec.ERC20Reference,
    block: spec.BlockNumberReference = 'latest',
    **rpc_kwargs
) -> int:
    """get decimals of an erc20"""
    return await erc20_generic.async_erc20_eth_call(
        function_name='decimals', token=token, block=block, **rpc_kwargs
    )


async def async_get_erc20s_decimals(
    tokens: typing.Iterable[spec.ERC20Reference],
    block: spec.BlockNumberReference = 'latest',
    **rpc_kwargs
) -> list[int]:
    """get decimals of multiple erc20s"""
    return await erc20_generic.async_erc20s_eth_calls(
        function_name='decimals', tokens=tokens, block=block, **rpc_kwargs
    )


async def async_get_erc20_decimals_by_block(
    token: spec.ERC20Reference, blocks=None, **rpc_kwargs
) -> list[int]:
    """get decimals of an erc20 across multiple blocks"""
    return await erc20_generic.async_erc20_eth_call_by_block(
        function_name='decimals', token=token, blocks=blocks, **rpc_kwargs
    )


#
# # name
#


async def async_get_erc20_name(
    token: spec.ERC20Reference,
    block: spec.BlockNumberReference = 'latest',
    **rpc_kwargs
) -> str:
    """get name of an erc20"""
    return await erc20_generic.async_erc20_eth_call(
        function_name='name', token=token, block=block, **rpc_kwargs
    )


async def async_get_erc20s_names(
    tokens: typing.Iterable[spec.ERC20Reference],
    block: spec.BlockNumberReference = 'latest',
    **rpc_kwargs
) -> list[str]:
    """get name of multiple erc20s"""
    return await erc20_generic.async_erc20s_eth_calls(
        function_name='name', tokens=tokens, block=block, **rpc_kwargs
    )


async def async_get_erc20_name_by_block(
    token: spec.ERC20Reference,
    blocks: typing.Iterable[spec.BlockNumberReference],
    **rpc_kwargs
) -> list[str]:
    """get name of an erc20 across multiple blocks"""
    return await erc20_generic.async_erc20_eth_call_by_block(
        function_name='name', token=token, blocks=blocks, **rpc_kwargs
    )


#
# # symbol
#


async def async_get_erc20_symbol(
    token: spec.ERC20Reference,
    block: spec.BlockNumberReference = 'latest',
    **rpc_kwargs
):
    """get symbol of an erc20"""
    return await erc20_generic.async_erc20_eth_call(
        function_name='symbol', token=token, block=block, **rpc_kwargs
    )


async def async_get_erc20s_symbols(
    tokens: typing.Iterable[spec.ERC20Reference],
    block: spec.BlockNumberReference = 'latest',
    **rpc_kwargs
):
    """get symbol of multiple erc20s"""
    return await erc20_generic.async_erc20s_eth_calls(
        function_name='symbol', tokens=tokens, block=block, **rpc_kwargs
    )


async def async_get_erc20_symbol_by_block(
    token: spec.ERC20Reference,
    blocks: typing.Iterable[spec.BlockNumberReference],
    **rpc_kwargs
):
    """get symbol of an erc20 across multiple blocks"""
    return await erc20_generic.async_erc20_eth_call_by_block(
        function_name='symbol', token=token, blocks=blocks, **rpc_kwargs
    )




from __future__ import annotations

import asyncio

from . import analytics_spec
from . import metric_groups


f_metric_group: dict[str, analytics_spec.MetricGroupCreatorCoroutine] = {
    'pcv_by_asset': metric_groups.async_compute_pcv_by_asset,
    'prices': metric_groups.async_compute_prices,
    'dex_volume': metric_groups.async_compute_dex_volume,
    'dex_tvls': metric_groups.async_compute_dex_tvls,
    'pfei_by_platform': metric_groups.async_compute_pfei_by_platform,
    'buybacks': metric_groups.async_compute_buybacks,
}
f_multi_metric_group: dict[
    str, analytics_spec.MultiMetricGroupCreatorCoroutine
] = {
    'pcv_stats': metric_groups.async_compute_pcv_stats,
}


async def async_get_metrics(
    blocks: list[int],
    verbose: bool = False,
) -> dict[str, analytics_spec.MetricGroup]:

    # create coroutines
    mg_coroutines = [f(blocks, verbose) for f in f_metric_group.values()]
    mg_task = asyncio.gather(*mg_coroutines)
    mmg_coroutines = [f(blocks, verbose) for f in f_multi_metric_group.values()]
    mmg_task = asyncio.gather(*mmg_coroutines)

    # build payload
    payload = {}
    mg_results = await mg_task
    for key, mg_result in zip(f_metric_group.keys(), mg_results):
        payload[key] = mg_result
    mmg_results = await mmg_task
    for mmg_result in mmg_results:
        for key, value in mmg_result.items():
            payload[key] = value

    # add missing orders
    for key, metric_group in payload.items():
        if 'order' not in metric_group:
            metric_group['order'] = order_metrics_in_group(
                metric_group['metrics']
            )

    return payload


def order_metrics_in_group(
    metrics: dict[str, analytics_spec.MetricData]
) -> list[str]:
    """order metrics in descending order by their final values"""
    pairs = [
        (metric_data['values'][-1], metric_name)
        for metric_name, metric_data in metrics.items()
    ]
    pairs = sorted(pairs, reverse=True)
    return [pair[-1] for pair in pairs]




from ctc import spec
from ctc import binary
from .. import rpc_format
from .. import rpc_spec


def digest_eth_get_transaction_count(
    response: spec.RpcSingularResponse, decode_response: bool = True
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_get_transaction_by_hash(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        quantities = rpc_spec.rpc_transaction_quantities
        response = rpc_format.decode_response(response, quantities)
    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)
    return response


def digest_eth_get_transaction_by_block_hash_and_index(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        quantities = rpc_spec.rpc_transaction_quantities
        response = rpc_format.decode_response(response, quantities)
    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)
    return response


def digest_eth_get_transaction_by_block_number_and_index(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        quantities = rpc_spec.rpc_transaction_quantities
        response = rpc_format.decode_response(response, quantities)
    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)
    return response


def digest_eth_get_transaction_receipt(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        quantities = rpc_spec.rpc_transaction_receipt_quantities
        response = rpc_format.decode_response(response, quantities)
    if snake_case_response:
        response = rpc_format.keys_to_snake_case(response)
    return response


def digest_eth_get_block_transaction_count_by_hash(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_get_block_transaction_count_by_number(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response




"""
digestors are simpler than constructors and so they do not have batch versions
"""

from __future__ import annotations

import inspect
import typing

from ctc import spec
from .. import rpc_registry
from .. import rpc_request
from .. import rpc_spec


#
# # batch construction
#


def batch_construct(method: str, **constructor_kwargs) -> spec.RpcPluralRequest:
    """construct a batch of rpc calls"""
    batch_inputs = _get_batch_constructor_inputs(method=method)
    if len(batch_inputs) == 0:
        raise Exception('no batch inputs available for method: ' + str(method))
    singular_constructor = rpc_registry.get_constructor(method=method)
    parameter, values, constructor_kwargs = _get_batch_parameter(
        constructor_kwargs,
        batch_inputs,
    )
    return [
        singular_constructor(**{parameter: value}, **constructor_kwargs)
        for value in values
    ]


def _get_batch_parameter(kwargs, batch_inputs):
    """identify the batch parameter given in kwargs

    return (singular_parameter, parameter_value, other_kwargs)
    """

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    # find suitable candidates
    candidates = []
    for plural_name, singular_name in batch_inputs.items():
        if plural_name in kwargs:
            candidates.append(plural_name)

    # select candidate
    if len(candidates) == 0:
        raise Exception('no batch parameter specified')
    elif len(candidates) > 1:
        raise Exception('too many batch parameters specified')
    else:
        parameter = candidates[0]
        return (
            batch_inputs[parameter],
            kwargs[parameter],
            {k: v for k, v in kwargs.items() if k not in batch_inputs},
        )


def _get_batch_constructor_inputs(method: str) -> dict[str, str]:
    return rpc_spec.rpc_constructor_batch_inputs.get(method, {})


#
# # batch execution
#


def batch_execute(
    method: str, *, provider: spec.ProviderSpec = None, **kwargs
) -> spec.RpcPluralResponse:
    """execute batch rpc call synchronously"""

    constructor_kwargs, digestor_kwargs = _separate_execution_kwargs(
        method=method,
        kwargs=kwargs,
    )
    request = batch_construct(method=method, **constructor_kwargs)
    response = rpc_request.send(request=request, provider=provider)
    return batch_digest(response=response, method=method, **digestor_kwargs)


async def async_batch_execute(
    method: str, *, provider: spec.ProviderSpec = None, **kwargs
) -> spec.RpcPluralResponse:
    """execute batch rpc call asynchronously"""

    constructor_kwargs, digestor_kwargs = _separate_execution_kwargs(
        method=method,
        kwargs=kwargs,
    )
    request = batch_construct(method=method, **constructor_kwargs)
    response = await rpc_request.async_send(request=request, provider=provider)
    return batch_digest(response=response, method=method, **digestor_kwargs)


def _separate_execution_kwargs(
    method: str,
    kwargs: dict[str, typing.Any],
) -> tuple[dict[str, typing.Any], dict[str, typing.Any]]:
    """separate constructor kwargs from digestor kwargs"""

    # compile digestor kwargs
    digestor = rpc_registry.get_digestor(method)
    signature = inspect.getfullargspec(digestor)
    digestor_args = signature.args + signature.kwonlyargs

    # separate kwargs into constructor and digestor kwargs
    constructor_kwargs = {}
    digestor_kwargs = {}
    for key, value in kwargs.items():
        if key in digestor_args:
            digestor_kwargs[key] = value
        else:
            constructor_kwargs[key] = value

    # add args that are passed to both constructors and digestors
    if method == 'eth_call':
        if kwargs.get('function_abi') is not None:
            constructor_kwargs['function_abi'] = kwargs['function_abi']

    return constructor_kwargs, digestor_kwargs


#
# # batch digestion
#


def batch_digest(
    response: spec.RpcPluralResponse, method: str, **digestor_kwargs
) -> spec.RpcPluralResponse:

    digestor = rpc_registry.get_digestor(method)
    results = []
    for s, subresponse in enumerate(response):
        result = digestor(subresponse, **digestor_kwargs)
        results.append(result)
    return results




import asyncio

from ctc import binary
from ctc import directory
from ctc import rpc
from ctc import spec
from ... import abi_utils
from ... import block_utils


async def async_get_events_from_node(
    start_block='latest',
    end_block='latest',
    event_name=None,
    event_hash=None,
    event_abi=None,
    contract_address=None,
    contract_abi=None,
    blocks_per_chunk=2000,
    package_as_dataframe=True,
    verbose=True,
    provider: spec.ProviderSpec = None,
):
    """see fetch_events() for complete kwarg list"""

    provider = rpc.get_provider(provider)
    network = provider['network']
    if network is None:
        raise Exception('could not determine network')

    # create chunks
    start_block, end_block = await block_utils.async_block_numbers_to_int(
        blocks=[start_block, end_block],
        provider=provider,
    )
    if verbose:
        print(
            'getting events from node, block range:', [start_block, end_block]
        )
    blocks_per_chunk = min(1000, blocks_per_chunk)
    if blocks_per_chunk is not None:
        chunks = block_utils.get_chunks_in_range(
            start_block=start_block,
            end_block=end_block,
            chunk_size=blocks_per_chunk,
            trim_excess=True,
        )
    else:
        chunks = [[start_block, end_block]]

    # gather metadata
    if contract_abi is None and event_abi is None:
        contract_abi = await abi_utils.async_get_contract_abi(
            contract_address=contract_address,
            network=network,
        )
    if event_name is None and event_hash is None and event_abi is None:
        raise Exception('must specify event_name or event_hash or event_abi')
    if event_name is None:
        if event_abi is None:
            event_abi = await abi_utils.async_get_event_abi(
                event_hash=event_hash,
                contract_abi=contract_abi,
                contract_address=contract_address,
                network=network,
            )
        event_name = event_abi['name']
    if event_hash is None:
        if event_abi is None:
            event_abi = await abi_utils.async_get_event_abi(
                event_hash=event_hash,
                contract_abi=contract_abi,
                contract_address=contract_address,
                network=network,
            )
        if event_name is not None:
            event_hash = binary.get_event_hash(event_abi=event_abi)
        elif event_abi is not None:
            event_hash = contract_address.get_event_hash(event_abi=event_abi)

        else:
            raise Exception('must specify event_name, event_abi, or event_hash')

    # fetch events
    coroutines = []
    for chunk in chunks:
        coroutine = _async_get_chunk_of_events_from_node(
            block_range=chunk,
            event_hash=event_hash,
            contract_address=contract_address,
            verbose=verbose,
            provider=provider,
        )
        coroutines.append(coroutine)
    chunks_entries = await asyncio.gather(*coroutines)
    entries = [
        entry for chunk_entries in chunks_entries for entry in chunk_entries
    ]

    # package as dataframe
    if package_as_dataframe:
        entries = await _async_package_exported_events(
            entries,
            contract_address=contract_address,
            contract_abi=contract_abi,
            event_hash=event_hash,
            event_name=event_name,
            event_abi=event_abi,
            provider=provider,
        )

    return entries


async def _async_get_chunk_of_events_from_node(
    block_range,
    event_hash,
    contract_address,
    verbose,
    provider=None,
):

    if verbose > 1:
        print('scraping block range: ' + str(block_range) + '\n', end='')

    # fetch entries
    start_block, end_block = block_range
    entries = await rpc.async_eth_get_logs(
        address=contract_address,
        topics=[event_hash],
        start_block=start_block,
        end_block=end_block,
        provider=provider,
    )

    return entries


async def _async_package_exported_events(
    entries,
    contract_address,
    contract_abi,
    event_hash,
    event_name,
    event_abi,
    provider,
):
    if event_abi is None:
        event_abi = await abi_utils.async_get_event_abi(
            contract_address=contract_address,
            contract_abi=contract_abi,
            event_hash=event_hash,
            event_name=event_name,
            provider=provider,
        )

    if len(entries) == 0:
        return create_empty_event_dataframe(event_abi=event_abi)

    formatted_entries = []
    for entry in entries:
        formatted_entry = binary.normalize_event(
            event=entry,
            event_abi=event_abi,
        )
        formatted_entries.append(formatted_entry)

    import pandas as pd

    df = pd.DataFrame(formatted_entries)
    df = df.set_index(['block_number', 'transaction_index', 'log_index'])

    return df


def create_empty_event_dataframe(
    event_abi=None,
):

    # standard columns
    columns = [
        'block_number',
        'transaction_index',
        'log_index',
        'block_hash',
        'transaction_hash',
        'contract_address',
        'event_name',
    ]

    # event-specific columns
    for item in event_abi['inputs']:
        columns.append('arg__' + item['name'])

    import pandas as pd

    df = pd.DataFrame(columns=columns)
    df = df.set_index(['block_number', 'transaction_index', 'log_index'])

    return df




from ctc import binary


def get_command_spec():
    return {
        'f': keccack_command,
        'args': [
            {'name': 'data'},
            {'name': '--text', 'kwargs': {'action': 'store_true'}},
            {'name': '--hex', 'kwargs': {'action': 'store_true'}},
        ],
    }


def keccack_command(data, text, hex):
    if text:
        hex = False
    elif hex:
        pass
    else:
        hex = data.startswith('0x')

    if hex:
        keccak = binary.keccak(data)
    else:
        keccak = binary.keccak_text(data)
    print(keccak)




from ctc import spec
from ctc import binary
from .. import rpc_format
from .. import rpc_spec


def digest_eth_new_filter(
    response: spec.RpcSingularResponse, decode_response: bool = False
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_new_block_filter(
    response: spec.RpcSingularResponse, decode_response: bool = False
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_new_pending_transaction_filter(
    response: spec.RpcSingularResponse, decode_response: bool = False
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_uninstall_filter(
    response: spec.RpcSingularResponse, decode_response: bool = False
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_get_filter_changes(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
    include_removed: bool = False,
) -> spec.RpcSingularResponse:
    if (
        not include_removed
        and len(response) > 0
        and isinstance(response[0], dict)
    ):
        response = [
            subresponse
            for subresponse in response
            if not subresponse['removed']
        ]

    if decode_response and len(response) > 0 and isinstance(response[0], dict):
        response = [
            rpc_format.decode_response(subresponse, rpc_spec.rpc_log_quantities)
            for subresponse in response
        ]

    if (
        snake_case_response
        and len(response) > 0
        and isinstance(response[0], dict)
    ):
        response = [
            rpc_format.keys_to_snake_case(subresponse)
            for subresponse in response
        ]

    return response


def digest_eth_get_filter_logs(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
    include_removed: bool = False,
) -> spec.RpcSingularResponse:

    if not include_removed:
        response = [
            subresponse
            for subresponse in response
            if not subresponse['removed']
        ]

    if decode_response and len(response) > 0 and isinstance(response[0], dict):
        response = [
            rpc_format.decode_response(subresponse, rpc_spec.rpc_log_quantities)
            for subresponse in response
        ]

    if (
        snake_case_response
        and len(response) > 0
        and isinstance(response[0], dict)
    ):
        response = [
            rpc_format.keys_to_snake_case(subresponse)
            for subresponse in response
        ]

    return response


def digest_eth_get_logs(
    response: spec.RpcSingularResponse,
    decode_response: bool = True,
    snake_case_response: bool = True,
    include_removed: bool = False,
) -> spec.RpcSingularResponse:
    if not include_removed:
        response = [
            subresponse
            for subresponse in response
            if not subresponse['removed']
        ]

    if decode_response and len(response) > 0 and isinstance(response[0], dict):
        response = [
            rpc_format.decode_response(subresponse, rpc_spec.rpc_log_quantities)
            for subresponse in response
        ]

    if (
        snake_case_response
        and len(response) > 0
        and isinstance(response[0], dict)
    ):
        response = [
            rpc_format.keys_to_snake_case(subresponse)
            for subresponse in response
        ]

    return response




from ctc import spec
from ctc import binary
from .. import rpc_format


def digest_web3_client_version(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_web3_sha3(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_net_version(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_net_peer_count(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_net_listening(
    response: spec.RpcSingularResponse,
) -> spec.RpcSingularResponse:
    return response


def digest_eth_protocol_version(
    response: spec.RpcSingularResponse, decode_response: bool = True
) -> spec.RpcSingularResponse:
    if decode_response:
        response = binary.convert(response, 'integer')
    return response


def digest_eth_syncing(
    response: spec.RpcSingularResponse, snake_case_response: bool = True
) -> spec.RpcSingularResponse:
    if snake_case_response:
        if isinstance(response, dict):
            response = rpc_format.keys_to_snake_case(response)
    return response




from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    from typing_extensions import TypeGuard

from ..typedefs import binary_types


def is_data(data: typing.Any) -> TypeGuard[binary_types.Data]:
    return (
        is_integer_data(data)
        or is_binary_data(data)
        or is_hex_data(data)
    )


def is_integer_data(data: typing.Any) -> TypeGuard[binary_types.IntegerData]:
    return isinstance(data, int)


def is_binary_data(data: typing.Any) -> TypeGuard[binary_types.BinaryData]:
    return isinstance(data, bytes)


def is_prefix_hex_data(
    data: typing.Any,
) -> TypeGuard[binary_types.PrefixHexData]:
    if not isinstance(data, str):
        return False
    if not data.startswith('0x'):
        return False
    try:
        int(data, 16)
        return True
    except ValueError:
        return False


def is_raw_hex_data(data: typing.Any) -> TypeGuard[binary_types.RawHexData]:
    if not isinstance(data, str):
        return False
    if data.startswith('0x'):
        return False
    try:
        int(data, 16)
        return True
    except ValueError:
        return False


def is_hex_data(data: typing.Any) -> TypeGuard[binary_types.HexData]:
    if not isinstance(data, str):
        return False
    try:
        int(data, 16)
        return True
    except ValueError:
        return False





usd_token = '0x1111111111111111111111111111111111111111'


coracle_addresses = {
    'StaticPCVDepositWrapper': '0x8B41DcEfAe6064E6bc2A9B3ae20141d23EFD6cbd',
    'CollateralizationOracle': '0xFF6f59333cfD8f4Ebc14aD0a0E181a83e655d257',
    'CollateralizationOracleWrapper_TransparentUpgradeableProxy': '0xd1866289B4Bd22D453fFF676760961e0898EE9BF',
    'CollateralizationOracleWrapper_ProxyImplementation': '0x656aA9c9875eB089b11869d4730d6963D25E76ad',
    'ProxyAdmin': '0xf8c2b645988b7658e7748ba637fe25bdd46a704a',
}
coracle_addresses = {k: v.lower() for k, v in coracle_addresses.items()}

token_aliases = {
    'WETH': 'ETH',
    'USD': 'Misc',
}
skip_tokens = [
    'CREAM',
]

deposit_metadata = {
    '0xb4ffd10c4c290dc13e8e30bf186f1509001515fd': {
        'name': 'Fuse Barnbridge Pool 25',
        'platform': 'Rari Fuse',
    },
    '0x4a5af5a124e672c156241b76cad4e41d09dd4883': {
        'name': 'Fuse Forex Pool 72',
        'platform': 'Rari Fuse',
    },
    '0x4e119714f625b2e82e5fb5a7e297978f020ea51e': {
        'name': 'Fuse G-UNI Pool 28',
        'platform': 'Rari Fuse',
    },
    '0x05e2e93cfb0b53d36a3151ee727bb581d4b918ce': {
        'name': 'Fuse NFTX Pool 31',
        'platform': 'Rari Fuse',
    },
    '0xd6598a23418c7fef7c0dc863265515b623b720f9': {
        'name': 'Fuse Fei Pool 8',
        'platform': 'Rari Fuse',
    },
    '0x76dfcf06e7d7b8248094dc319b284fb244f06309': {
        'name': 'Fuse Fei Pool 79',
        'platform': 'Rari Fuse',
    },
    '0x81dcb06ea4db474d1506ca6275ff7d870ba3a1be': {
        'name': 'Fuse NFTX Pool 31',
        'platform': 'Rari Fuse',
    },
    '0xb13c755107301ebfed6a93190acde09281b2f8a5': {
        'name': 'Fuse UpOnly Pool 7',
        'platform': 'Rari Fuse',
    },
    '0x07f2dd7e6a78d96c08d0a8212f4097dcc129d629': {
        'name': 'Fuse Ohm Pool 18',
        'platform': 'Rari Fuse',
    },
    '0xe2e35097638f0ff2eeca2ef70f352be37431945f': {
        'name': 'Fuse StakeDAO Pool 27',
        'platform': 'Rari Fuse',
    },
    '0xb0d5eba35e1cece568096064ed68a49c6a24d961': {
        'name': 'Fuse G-UNI Pool 28',
        'platform': 'Rari Fuse',
    },
    '0x7aa4b1558c3e219cfffd6a356421c071f71966e7': {
        'name': 'Fuse Tetranode\'s Locker Pool 6',
        'platform': 'Rari Fuse',
    },
    '0x2296a2417d1f02d394ab22af794a0f426ed53436': {
        'name': 'Fuse Liquity Pool 91',
        'platform': 'Rari Fuse',
    },
    # '0xff419bc27483edb94b7ad5c97b7fab5db323c7e0': {
    #     'name': 'CREAM',
    #     'platform': 'CREAM',
    # },
    '0x7e39bba9d0d967ee55524fae9e54900b02d9889a': {
        'name': 'Fuse Index Coop Pool 19',
        'platform': 'Rari Fuse',
    },
    '0x96a657ee40a79a964c6b4ea551c895d98e885a75': {
        'name': 'Fuse Stable Asset Pool 9',
        'platform': 'Rari Fuse',
    },
    '0x7eb88140af813294aedce981b6ac08fcd139d408': {
        'name': 'OA Account',
        'platform': 'OA Account',
    },
    '0x508f6fbd78b6569c29e9d75986a51558de9e5865': {
        'name': 'Fuse Harvest Pool 24',
        'platform': 'Rari Fuse',
    },
    '0x82aebee64a52180d8541eb601a8381e012a1ed04': {
        'name': 'Fuse Tokemak Pool 26',
        'platform': 'Rari Fuse',
    },
    '0x61d26126d2f8a44b41c1d8e1b1f276551dc8eec6': {
        'name': 'Fuse Float Pool 90',
        'platform': 'Rari Fuse',
    },
    '0x107460564896377ba6cdcc7516c7eab65e32e360': {
        'name': 'Balancer FEI-TRIBE LBP',
        'platform': 'Balancer',
    },
    '0xfac571b6054619053ac311da8112939c9a374a85': {
        'name': 'Aave V2 FEI Pool',
        'platform': 'Aave',
    },
    '0x15958381e9e6dc98bd49655e36f524d2203a28bd': {
        'name': 'Uniswap FEI-ETH',
        'platform': 'Uniswap V2',
    },
    '0x5ae217de26f6ff5f481c6e10ec48b2cf2fc857c8': {
        'name': 'Curve d3',
        'platform': 'Curve',
    },
    '0x902199755219a9f8209862d09f1891cfb34f59a3': {
        'name': 'Sushiswap FEI-DPI',
        'platform': 'Sushi',
    },
    '0x1f05b337cb16cea2a1c638ba9b9571f0cf4a5612': {
        'name': 'Balancer FEI-USD',
        'platform': 'Balancer',
    },
    '0x7ac2ab8143634419c5bc230a9f9955c3e29f64ef': {
        'name': 'Uniswap FEI-agEUR',
        'platform': 'Uniswap V2',
    },
}


def get_coracle_address(wrapper=False, block=None, blocks=None):

    if blocks is not None:
        return [
            get_coracle_address(wrapper=wrapper, block=block)
            for block in blocks
        ]

    if wrapper:
        return coracle_addresses[
            'CollateralizationOracleWrapper_TransparentUpgradeableProxy'
        ]
    else:
        return coracle_addresses['CollateralizationOracle']




import os

import ctc.config


def get_command_spec():
    return {
        'f': config_command,
    }


def config_command():

    env_var = ctc.config.config_path_env_var

    print('# Config Summary')
    print('- config env variable:', env_var)
    if env_var not in os.environ:
        print('-', env_var, 'not set')
    else:
        env_value = os.environ[env_var]
        if env_value is None or env_value == '':
            print('-', env_var, 'set to null')
        else:
            print('-', env_var, 'set to:', env_value)
    print('- config path:', ctc.config.get_config_path(raise_if_dne=False))

    print()
    print('## Config Values')
    config = ctc.config.get_config()
    for key in sorted(config.keys()):
        if isinstance(config[key], dict) and len(config[key]) > 0:
            print('-', str(key) + ':')
            for subkey, subvalue in config[key].items():
                if isinstance(subvalue, dict) and len(subvalue) > 0:
                    print('    -', str(subkey) + ':')
                    for subsubkey, subsubvalue in subvalue.items():
                        print('        -', str(subsubkey) + ':', subsubvalue)
                else:
                    print('    -', str(subkey) + ':', subvalue)
        else:
            print('-', str(key) + ':', config[key])




from .abi_types import *
from .address_types import *
from .binary_types import *
from .block_types import *
from .config_types import *
from .external_types import *
from .network_types import *
from .rpc_types import *
from .storage_types import *



import asyncio
import functools

from ctc.toolbox import search_utils
from .. import address_utils
from . import block_crud


async def async_get_contract_creation_block(
    contract_address,
    start_block=None,
    end_block=None,
    provider=None,
    verbose=True,
    nary=None,
):
    """get the block where a contract was created

    algorithm: perform a binary search across blocks, check code bytes in each
    """

    contract_address = address_utils.get_address_checksum(contract_address)

    if start_block is None:
        start_block = 0
    if end_block is None:
        end_block = 'latest'
    if start_block == 'latest' or end_block == 'latest':
        latest_block = await block_crud.async_get_latest_block_number()
        if start_block == 'latest':
            start_block = latest_block
        if end_block == 'latest':
            end_block = latest_block

    if verbose:
        print('searching for creation block of ' + contract_address)

    async def async_is_match(index):
        if verbose:
            print('- trying block:', index)
        return await address_utils.async_is_contract_address(
            address=contract_address, block=index, provider=provider
        )

    if nary is None:
        result = await search_utils.async_binary_search(
            start_index=start_block,
            end_index=end_block,
            async_is_match=async_is_match,
        )
    else:
        result = search_utils.nary_search(
            start_index=start_block,
            end_index=end_block,
            is_match=is_match,
            nary=nary,
        )

    if verbose:
        print('result:', result)

    return result


async def async_get_blocks_of_timestamps(
    timestamps,
    block_timestamps=None,
    block_number_array=None,
    block_timestamp_array=None,
    nary=None,
    cache=None,
    provider=None,
):
    """once parallel node search created, use that"""

    if (
        block_timestamps is not None
        and block_number_array is None
        and block_timestamp_array is None
    ):
        import numpy as np

        block_timestamp_array = np.array(list(block_timestamps.values()))
        block_number_array = np.array(list(block_timestamps.keys()))

        blocks = []
        for timestamp in timestamps:
            block = get_block_of_timestamp_from_arrays(
                timestamp=timestamp,
                nary=nary,
                cache=cache,
                block_timestamps=block_timestamps,
                block_timestamp_array=block_timestamp_array,
                block_number_array=block_number_array,
            )
            blocks.append(block)

        return blocks

    else:

        coroutines = []
        for timestamp in timestamps:
            coroutine = async_get_block_of_timestamp(
                timestamp=timestamp,
                verbose=False,
            )
            coroutines.append(coroutine)

        return await asyncio.gather(*coroutines)


# @toolparallel.parallelize_input(
#     singular_arg='timestamp', plural_arg='timestamps'
# )
async def async_get_block_of_timestamp(
    timestamp,
    nary=None,
    cache=None,
    block_timestamps=None,
    block_timestamp_array=None,
    block_number_array=None,
    verbose: bool = True,
    provider=None,
):
    if (
        block_timestamps is not None
        or block_timestamp_array is not None
        and block_number_array is not None
    ):
        return get_block_of_timestamp_from_arrays(
            timestamp=timestamp,
            block_timestamp_array=block_timestamp_array,
            block_number_array=block_number_array,
            verbose=verbose,
        )
    else:
        return await async_get_block_of_timestamp_from_node(
            timestamp=timestamp,
            nary=nary,
            cache=cache,
            verbose=verbose,
        )


def get_block_of_timestamp_from_arrays(
    timestamp, block_timestamp_array, block_number_array, verbose
):
    import numpy as np

    index = np.searchsorted(block_timestamp_array, timestamp)
    return block_number_array[index]


async def async_get_block_of_timestamp_from_node(
    timestamp, nary=None, cache=None, provider=None, verbose: bool = True
):
    """
    - could make this efficiently parallelizable to multiple timestamps by sharing cache
        - would need to remove the initializing key from the shared cache
    """

    if nary is None:
        nary = 6

    if cache is None:
        cache = {'initializing': {timestamp: True}, 'timestamps': {}}

    async_is_match = functools.partial(
        _async_is_match_block_of_timestamp, timestamp=timestamp, cache=cache
    )
    get_next_probes = functools.partial(
        _get_next_probes_block_of_timestamp,
        timestamp=timestamp,
        cache=cache,
        debug=verbose,
    )

    return await search_utils.async_nary_search(
        nary=nary,
        start_index=1,
        end_index=(await block_crud.async_get_latest_block_number()),
        async_is_match=async_is_match,
        get_next_probes=get_next_probes,
    )


async def _async_is_match_block_of_timestamp(
    block_numbers, timestamp, cache, provider=None
):

    # retrieve values not in cache
    not_in_cache = [
        block_number
        for block_number in block_numbers
        if block_number not in cache['timestamps']
    ]
    gotten = await block_crud.async_get_blocks(not_in_cache)
    for block_number, block in zip(not_in_cache, gotten):
        cache['timestamps'][block_number] = block['timestamp']

    # compute results
    return [
        cache['timestamps'][block_number] >= timestamp
        for block_number in block_numbers
    ]


def _get_next_probes_block_of_timestamp(
    nary, probe_min, probe_max, timestamp, cache, debug=True
):
    import numpy as np

    if cache['initializing'][timestamp]:
        cache['initializing'][timestamp] = False
        return search_utils.get_next_probes_linear(probe_min, probe_max, nary)
    else:

        min_timestamp = cache['timestamps'][probe_min]
        total_time = cache['timestamps'][probe_max] - min_timestamp
        total_blocks = probe_max - probe_min
        mean_block_time = total_time / total_blocks
        target_index = probe_min + (timestamp - min_timestamp) / mean_block_time
        target_index = int(target_index)

        if debug:
            print('mean_block_time:', mean_block_time)
            print('probe_min:', probe_min)
            print('min_timestamp:', min_timestamp)
            print('target_index:', target_index)
            print()

        # if narrowing a range larger than N, probe a window of size X centered on target_index
        if probe_max - probe_min > 1000:
            size = probe_max - probe_min
            probes = target_index + size * np.linspace(-0.01, 0.01, nary - 1)
            probes = probes.astype(int)
            probes = [
                probe
                for probe in probes
                if probe <= probe_max and probe >= probe_min
            ]
            return probes

        # else, probe the indices directly surrounding target_index
        else:

            if target_index == probe_max:
                target_index -= 1
            elif target_index == probe_min:
                target_index += 1

            n_probes = nary - 1
            half = int(n_probes / 2)
            probes = [target_index + i for i in range(-half, n_probes - half)]

            return probes




from . import backend_exceptions


def get_backend_order(backend, backend_order):
    if backend_order is None and backend is None:
        backend_order = ['filesystem', 'node']
    elif backend is not None:
        backend_order = [backend]
    elif backend_order is not None:
        pass
    else:
        raise Exception('specify backend or backend_order')
    return backend_order


def run_on_backend(
    backend_functions, backend=None, backend_order=None, **function_kwargs
):
    backend_order = get_backend_order(backend, backend_order)
    for backend in backend_order:
        try:
            function = backend_functions.get(backend)
            return function(**function_kwargs)
        except backend_exceptions.DataNotFound:
            pass
    else:
        raise Exception('could not execute any of: ' + str(backend_functions))


async def async_run_on_backend(
    backend_functions, backend=None, backend_order=None, **function_kwargs
):
    backend_order = get_backend_order(backend, backend_order)
    for backend in backend_order:
        try:
            function = backend_functions.get(backend)
            return await function(**function_kwargs)
        except backend_exceptions.DataNotFound:
            pass
    else:
        raise Exception('could not execute any of: ' + str(backend_functions))


def transfer_backends(
    get,
    save,
    from_backend,
    to_backend,
    get_kwargs=None,
    save_kwargs=None,
    common_kwargs=None,
    **more_common_kwargs
):
    if common_kwargs is None:
        common_kwargs = {}
    common_kwargs = dict(common_kwargs, **more_common_kwargs)

    if get_kwargs is None:
        get_kwargs = {}
    get_kwargs = dict(common_kwargs, **get_kwargs)
    result = get(backend=from_backend, **get_kwargs)

    if save_kwargs is None:
        save_kwargs = {}
    save_kwargs = dict(common_kwargs, **get_kwargs)

    return save(result, backend=to_backend, **save_kwargs)


async def async_transfer_backends(
    get,
    save,
    from_backend,
    to_backend,
    get_kwargs=None,
    save_kwargs=None,
    common_kwargs=None,
    **more_common_kwargs
):
    if common_kwargs is None:
        common_kwargs = {}
    common_kwargs = dict(common_kwargs, **more_common_kwargs)

    if get_kwargs is None:
        get_kwargs = {}
    get_kwargs = dict(common_kwargs, **get_kwargs)
    result = await get(backend=from_backend, **get_kwargs)

    if save_kwargs is None:
        save_kwargs = {}
    save_kwargs = dict(common_kwargs, **get_kwargs)

    return await save(result, backend=to_backend, **save_kwargs)

